# Generated by Django 3.2.5 on 2023-03-28 08:50

import apps.accounts.fields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="UserProfile",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("info", models.TextField(blank=True, default="{}")),
                ("usage", models.TextField(blank=True, default="{}")),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="profile",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Site",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(help_text="Name of Mercury Site", max_length=256),
                ),
                (
                    "slug",
                    models.CharField(
                        help_text="Subdomain", max_length=256, unique=True
                    ),
                ),
                (
                    "domain",
                    models.CharField(
                        blank=True,
                        default="runmercury.com",
                        help_text="Domain address",
                        max_length=256,
                        null=True,
                    ),
                ),
                (
                    "custom_domain",
                    models.CharField(
                        blank=True,
                        help_text="Custom domain address",
                        max_length=256,
                        null=True,
                    ),
                ),
                (
                    "share",
                    models.CharField(
                        choices=[
                            ("PUBLIC", "Anyone can access notebooks and execute"),
                            ("PRIVATE", "Only selected users have access to notebooks"),
                        ],
                        default="PUBLIC",
                        max_length=32,
                    ),
                ),
                ("status", models.CharField(default="Ready", max_length=32)),
                ("welcome", models.TextField(blank=True, default="", null=True)),
                ("info", models.TextField(blank=True, null=True)),
                ("active", models.BooleanField(default=True)),
                (
                    "created_at",
                    apps.accounts.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "updated_at",
                    apps.accounts.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Secret",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=256)),
                ("token", models.TextField()),
                (
                    "created_at",
                    apps.accounts.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "hosted_on",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="accounts.site"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "rights",
                    models.CharField(
                        choices=[
                            ("VIEW", "View and execute notebooks"),
                            ("EDIT", "Edit and view site, files and execute notebooks"),
                        ],
                        default="VIEW",
                        help_text="Rights for user",
                        max_length=32,
                    ),
                ),
                (
                    "created_at",
                    apps.accounts.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "updated_at",
                    apps.accounts.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="created_by",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "host",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="hosts",
                        to="accounts.site",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Invitation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("invited", models.CharField(max_length=256)),
                (
                    "created_at",
                    apps.accounts.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "rights",
                    models.CharField(
                        choices=[
                            ("VIEW", "View and execute notebooks"),
                            ("EDIT", "Edit and view site, files and execute notebooks"),
                        ],
                        default="VIEW",
                        help_text="Rights for user",
                        max_length=32,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "hosted_on",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="accounts.site"
                    ),
                ),
            ],
        ),
    ]



import copy

from nbconvert import HTMLExporter, SlidesExporter


class Exporter:
    def __init__(
        self,
        show_code=False,
        show_prompt=False,
        is_presentation=False,
        reveal_theme="white",
    ):
        self.show_code = show_code
        self.show_prompt = show_prompt
        self.is_presentation = is_presentation
        self.reveal_theme = reveal_theme

        self.set_exporter()

    def set_exporter(self):
        if self.is_presentation:
            self.html_exporter = SlidesExporter(
                template_name="reveal", reveal_theme=self.reveal_theme
            )
        else:
            self.html_exporter = HTMLExporter(template_name="lab")

        # prompt
        self.html_exporter.exclude_input_prompt = not self.show_prompt
        self.html_exporter.exclude_output_prompt = not self.show_prompt

        # hide code
        self.html_exporter.exclude_input = not self.show_code
        self.html_exporter.exclude_input_prompt = not self.show_code
        self.html_exporter.exclude_output_prompt = not self.show_code

    def set_is_presentation(self, new_value):
        if self.is_presentation != new_value:
            self.is_presentation = new_value
            self.set_exporter()

    def set_reveal_theme(self, new_value):
        if self.reveal_theme != new_value:
            self.reveal_theme = new_value
            self.set_exporter()

    def set_show_code(self, new_value):
        self.show_code = new_value
        self.html_exporter.exclude_input = not self.show_code
        self.html_exporter.exclude_input_prompt = not self.show_code
        self.html_exporter.exclude_output_prompt = not self.show_code

    def set_show_prompt(self, new_value):
        self.show_prompt = new_value
        self.html_exporter.exclude_input_prompt = not self.show_prompt
        self.html_exporter.exclude_output_prompt = not self.show_prompt

    def is_stop_execution_output(self, output):
        return output.get(
            "output_type", ""
        ) == "error" and "StopExecution" in output.get("ename", "")

    def export(self, notebook, full_header=True):
        #
        # notebook needs to be in nbformat
        #
        n = copy.deepcopy(notebook)

        # omit cells with "---" (starter from V1)
        n.cells = [
            cell
            for cell in n.cells
            if not "".join(cell.get("source", "")).startswith("---")
        ]

        # omit output with mercury widgets
        for cell in n.cells:
            if "outputs" in cell:
                cell["outputs"] = [
                    output
                    for output in cell["outputs"]
                    if not "application/mercury+json" in output.get("data", {})
                ]
        # omit output with mercury StopExecution
        for cell in n.cells:
            if "outputs" in cell:
                cell["outputs"] = [
                    output
                    for output in cell["outputs"]
                    if not self.is_stop_execution_output(output)
                ]

        body, _ = self.html_exporter.from_notebook_node(n)

        # remove header
        if not full_header:
            body = self.remove_header(body)

        return body

    def remove_header(self, nb_body):
        body = copy.deepcopy(nb_body)
        index_start = body.find("<head>")
        index_end = body.find("</head>")

        if index_start != -1 and index_end != -1:
            body = body[:index_start] + "" + body[index_end + 7 :]

            body = body.replace("<!DOCTYPE html>", "")
            body = body.replace("<html>", "")
            body = body.replace("<body ", "<div ")
            body = body.replace("</body>", "</div>")
            body = body.replace("</html>", "")
        return body



import os
import sys
from pathlib import Path

from dotenv import load_dotenv

# Please put your env variables in .env file in the main directory of your project.
# The main directory is the directory from where you call mercury executable
# or python manage.py script.
#
# The example content of .env file:
# DEBUG=False
# SERVE_STATIC=True
# SECRET_KEY=django-insecure-)$12ir6-s6vbcufpva*va7bf$s$$(76ue$twwz9noath0&e91h

load_dotenv(".env")
load_dotenv("../.env")


# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

FRONTEND_BUILD_DIR = str(BASE_DIR / "frontend-dist")
FRONTEND_STATIC_DIR = str(BASE_DIR / "frontend-dist" / "static")

STORAGE_MEDIA = "media"
STORAGE_S3 = "s3"
STORAGE = STORAGE_MEDIA
if os.environ.get("STORAGE", STORAGE_MEDIA) == STORAGE_S3:
    STORAGE = STORAGE_S3
DJANGO_DRF_FILEPOND_UPLOAD_TMP = str(BASE_DIR / "uploads-temp")
DJANGO_DRF_FILEPOND_FILE_STORE_PATH = str(BASE_DIR / "uploads")


CELERY_BROKER_URL = os.environ.get("CELERY_BROKER_URL", "sqla+sqlite:///celery.sqlite")
CELERY_RESULT_BACKEND = os.environ.get("CELERY_RESULT_BACKEND", "db+sqlite:///celery.sqlite")

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get(
    "SECRET_KEY", "django-insecure-)$12ir6-s6vbcufpva*va7bf$s$$(76ue$twwz9noath0&e91h"
)
# Please keep SECRET_KEY secret!
# Generate new SECRET_KEY with the following code:
# from django.core.management.utils import get_random_secret_key
# print(get_random_secret_key())


# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get("DEBUG", "True") == "True"
SERVE_STATIC = os.environ.get("SERVE_STATIC", "False") == "True"

ALLOWED_HOSTS = [
    "api.docker",
    "server",
    "127.0.0.1",
    "0.0.0.0",
    "localhost",
    "mercury.mljar.com",
]

if os.environ.get("ALLOWED_HOSTS") is not None:
    try:
        ALLOWED_HOSTS += os.environ.get("ALLOWED_HOSTS").split(",")
    except Exception as e:
        print("Cant set ALLOWED_HOSTS, using default")

# Application definition

INSTALLED_APPS = [
    "daphne",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    # 3rd party
    "rest_framework",
    "corsheaders",
    "django_drf_filepond",
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    "dj_rest_auth",
    "dj_rest_auth.registration",
    "rest_framework.authtoken",
    # Mercury apps
    "apps.tasks",
    "apps.notebooks",
    "apps.ws",
    "apps.storage",
    "apps.accounts",
    "apps.workers",
]


REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.TokenAuthentication",
    )
}

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    "allauth.account.auth_backends.AuthenticationBackend",
)

REST_AUTH = {"USER_DETAILS_SERIALIZER": "apps.accounts.serializers.UserSerializer"}


SITE_ID = 1

ACCOUNT_EMAIL_REQUIRED = True
ACCOUNT_UNIQUE_EMAIL = True
ACCOUNT_USERNAME_REQUIRED = False
ACCOUNT_AUTHENTICATION_METHOD = "email"
ACCOUNT_EMAIL_VERIFICATION = os.environ.get("ACCOUNT_EMAIL_VERIFICATION", "mandatory")  # "none" or "mandatory"
ACCOUNT_EMAIL_CONFIRMATION_HMAC = False
OLD_PASSWORD_FIELD_ENABLED = True  # use old password when password change in the app

# CORS_ALLOWED_ORIGINS = ["http://localhost:3000", "http://single-site.localhost:3000",
#                 "http://127.0.0.1:3000"]

# CORS_ALLOWED_ORIGIN_REGEXES = [
#     r"^http://\w+\.localhost:3000$",
# ]

CORS_ALLOW_ALL_ORIGINS = True

MIDDLEWARE = ["django.middleware.security.SecurityMiddleware"]
if SERVE_STATIC:
    MIDDLEWARE += ["whitenoise.middleware.WhiteNoiseMiddleware"]

MIDDLEWARE += [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

X_FRAME_OPTIONS = "ALLOWALL"  # SAMEORGIN

ROOT_URLCONF = "server.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [FRONTEND_BUILD_DIR, BASE_DIR / 'templates'],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "server.wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DB_SQLITE = "sqlite"
DB_POSTGRESQL = "postgresql"

DATABASES_ALL = {
    DB_SQLITE: {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    },
    DB_POSTGRESQL: {
        "ENGINE": "django.db.backends.postgresql",
        "HOST": os.environ.get("POSTGRES_HOST", "localhost"),
        "NAME": os.environ.get("POSTGRES_NAME", "postgres"),
        "USER": os.environ.get("POSTGRES_USER", "postgres"),
        "PASSWORD": os.environ.get("POSTGRES_PASSWORD", "postgres"),
        "PORT": int(os.environ.get("POSTGRES_PORT", "5432")),
    },
}

DATABASES = {"default": DATABASES_ALL[os.environ.get("DJANGO_DB", DB_SQLITE)]}


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = os.environ.get("TIME_ZONE", "UTC")

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/
if DEBUG or SERVE_STATIC:
    STATIC_URL = "/static/"
    STATICFILES_DIRS = [FRONTEND_BUILD_DIR, FRONTEND_STATIC_DIR]
    if SERVE_STATIC:
        STATIC_ROOT = str(BASE_DIR / "static")
else:
    STATIC_URL = "/django_static/"
    STATIC_ROOT = str(BASE_DIR / "django_static")

MEDIA_ROOT = str(BASE_DIR / "media")

if not os.path.exists(MEDIA_ROOT):
    try:
        os.mkdir(MEDIA_ROOT)
    except Exception as e:
        raise Exception(f"Cannot create directory {MEDIA_ROOT}. {str(e)}")

MEDIA_URL = "/media/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


ASGI_APPLICATION = "server.asgi.application"

CHANNEL_LAYERS = {"default": {"BACKEND": "channels.layers.InMemoryChannelLayer"}}


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "DJ {levelname} {asctime} {module} {message}",
            "style": "{",
        },
        "simple": {
            "format": "DJ {levelname} {message}",
            "style": "{",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        },
        "file": {
            "level": "ERROR",
            "class": "logging.FileHandler",
            "formatter": "verbose",
            "filename": "django-errors.log",
        },
    },
    "loggers": {
        "django": {
            "handlers": ["console", "file"],
            "level": os.getenv("DJANGO_LOG_LEVEL", "ERROR"),
            "propagate": False,
        },
    },
}

NBWORKERS_PER_MACHINE = 20

# delta time after which worker is considered as stale and deleted
WORKER_STALE_TIME = 1  # in minutes

AWS_ACCESS_KEY_ID = os.environ.get("AWS_ACCESS_KEY_ID")
AWS_SECRET_ACCESS_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY")
AWS_REGION_NAME = os.environ.get("AWS_REGION_NAME")
AWS_BUCKET_NAME = os.environ.get("AWS_BUCKET_NAME")

# email setup, used in notification
EMAIL_HOST = os.environ.get("EMAIL_HOST")
EMAIL_HOST_USER = os.environ.get("EMAIL_HOST_USER")
EMAIL_HOST_PASSWORD = os.environ.get("EMAIL_HOST_PASSWORD")
EMAIL_PORT = os.environ.get("EMAIL_PORT")
EMAIL_USE_TLS = True  # use TLS by default
DEFAULT_FROM_EMAIL = os.environ.get("DEFAULT_FROM_EMAIL")

EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"
if os.environ.get("DJANGO_EMAIL_BACKEND", "console") == "smtp":
    EMAIL_BACKEND = "django.core.mail.backends.smtp.EmailBackend"
if os.environ.get("DJANGO_EMAIL_BACKEND", "console") == "ses":
    # you need to install django-ses
    EMAIL_BACKEND = "django_ses.SESBackend"
    AWS_SES_REGION_NAME = AWS_REGION_NAME
    AWS_SES_REGION_ENDPOINT = f'email.{AWS_REGION_NAME}.amazonaws.com'
    # DEFAULT_FROM_EMAIL = os.environ("DEFAULT_FROM_EMAIL", '"Mercury" <contact@runmercury.com>')



import os

from django.conf import settings
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.accounts.models import Site, Membership


class VersionInfo(APIView):
    def get(self, request, format=None):
        return Response({"isPro": True})


# it will be used as really simple cache
welcome_msg = None


class WelcomeMessage(APIView):
    def get(self, request, site_id, format=None):
        global welcome_msg
        welcome_file = os.environ.get("WELCOME")
        if welcome_file is not None:
            if welcome_msg is None:
                if os.path.exists(welcome_file):
                    with open(welcome_file, encoding="utf-8", errors="ignore") as fin:
                        welcome_msg = fin.read()
                else:
                    return Response({"msg": ""})
            if welcome_msg is not None:
                # check access rights ...
                user = request.user

                site = Site.objects.get(pk=site_id)
                if site.share == Site.PRIVATE:
                    if user.is_anonymous:
                        return Response({"msg": ""})
                    else:
                        # logged user needs to be owner or have member rights
                        owner = site.created_by == user
                        member = Membership.objects.filter(user=user, host=site)
                        if not owner and not member:
                            return Response({"msg": ""})

                return Response({"msg": welcome_msg})
        return Response({"msg": ""})



from django.conf.urls import url

from apps.notebooks.views import (
    GetNbIframes,
    ListNotebooks,
    RetrieveNotebook,
    RetrieveNotebookWithSlug,
)

notebooks_urlpatterns = [
    url(
        "api/v1/(?P<site_id>.+)/notebooks/(?P<notebook_id>.+)",
        RetrieveNotebook.as_view(),
    ),
    url(
        "api/v1/(?P<site_id>.+)/getnb/(?P<notebook_slug>.+)",
        RetrieveNotebookWithSlug.as_view(),
    ),
    url("api/v1/(?P<site_id>.+)/notebooks", ListNotebooks.as_view()),
    url("api/v1/(?P<site_id>.+)/nb-iframes", GetNbIframes.as_view()),
]



# Generated by Django 3.2.5 on 2023-03-29 09:06

import apps.accounts.fields
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('accounts', '0001_initial'),
        ('storage', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserUploadedFile',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('filename', models.CharField(max_length=1024)),
                ('filepath', models.CharField(max_length=1024)),
                ('session_id', models.CharField(max_length=128)),
                ('created_at', apps.accounts.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False)),
                ('hosted_on', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accounts.site')),
            ],
        ),
    ]



import json
import os
import shutil
import uuid

from celery.result import AsyncResult
from django.conf import settings
from django.db import transaction
from django.http import Http404
from django.shortcuts import get_object_or_404
from requests import request
from rest_framework import status
from rest_framework.exceptions import APIException
from rest_framework.generics import CreateAPIView, ListAPIView, RetrieveAPIView
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.notebooks.models import Notebook
from apps.notebooks.views import notebooks_queryset
from apps.storage.models import WorkerFile
from apps.storage.s3utils import S3
from apps.storage.storage import StorageManager
from apps.tasks.models import Task
from apps.tasks.serializers import TaskSerializer
from apps.tasks.tasks import task_execute
from apps.tasks.tasks_export import export_to_pdf


class TaskCreateView(CreateAPIView):
    serializer_class = TaskSerializer
    queryset = Task.objects.all()

    def perform_create(self, serializer):
        notebook = get_object_or_404(
            notebooks_queryset(self.request), pk=self.kwargs["notebook_id"]
        )
        try:
            with transaction.atomic():
                instance = serializer.save(
                    state="CREATED",
                    notebook=notebook,
                )
                job_params = {"db_id": instance.id}
                transaction.on_commit(lambda: task_execute.delay(job_params))
        except Exception as e:
            raise APIException(str(e))


class GetLastTaskView(RetrieveAPIView):
    serializer_class = TaskSerializer
    queryset = Task.objects.all()

    def get_object(self):
        notebook = get_object_or_404(
            notebooks_queryset(self.request), pk=self.kwargs["notebook_id"]
        )
        try:
            if notebook.schedule is None or notebook.schedule == "":
                return Task.objects.filter(
                    notebook_id=self.kwargs["notebook_id"],
                    session_id=self.kwargs["session_id"],
                ).latest("id")
            else:
                return Task.objects.filter(
                    notebook_id=self.kwargs["notebook_id"]
                ).latest("id")

        except Task.DoesNotExist:
            raise Http404()


class ListOutputFilesView(APIView):
    def get(self, request, session_id, task_id, format=None):
        files_urls = []
        try:
            output_dir = os.path.join(
                settings.MEDIA_ROOT, session_id, f"output_{task_id}"
            )
            for f in os.listdir(output_dir):
                if os.path.isfile(os.path.join(output_dir, f)):
                    files_urls += [
                        f"{settings.MEDIA_URL}/{session_id}/output_{task_id}/{f}"
                    ]
        except Exception as e:
            print(
                f"Trying to list files for session_id {session_id} and task_id {task_id}"
            )
            print("Exception occured", str(e))
        return Response(files_urls)


class ListWorkerOutputFilesView(APIView):
    def get(self, request, session_id, worker_id, notebook_id, format=None):
        files_urls = []
        sm = StorageManager(session_id, worker_id, notebook_id)
        if settings.STORAGE == settings.STORAGE_MEDIA:
            files_urls = sm.list_worker_files_urls()
        elif settings.STORAGE == settings.STORAGE_S3:
            output_dir = sm.worker_output_dir()
            files = WorkerFile.objects.filter(
                created_by__id=worker_id, output_dir=output_dir
            )
            s3 = S3()
            for f in files:
                files_urls += [s3.get_presigned_url(f.filepath, "get_object")]

        return Response(files_urls)


class ClearTasksView(APIView):
    def post(self, request, notebook_id, session_id, format=None):
        try:
            tasks = Task.objects.filter(notebook_id=notebook_id, session_id=session_id)

            for task in tasks:
                output_file = os.path.join(
                    settings.MEDIA_ROOT, session_id, f"output_{task.id}.html"
                )
                output_dir = os.path.join(
                    settings.MEDIA_ROOT, session_id, f"output_{task.id}"
                )

                try:
                    if os.path.isfile(output_file):
                        os.remove(output_file)
                    if os.path.isdir(output_dir):
                        shutil.rmtree(output_dir)
                except Exception as e:
                    print(f"Trying to delete {output_file} and {output_dir}")
                    print(str(e))

            tasks.delete()

        except Exception as e:
            print(
                f"Trying to clear tasks for notebook_id {notebook_id} and session_id {session_id}"
            )
            print("Exception occured", str(e))

        return Response(status=status.HTTP_204_NO_CONTENT)


class CreateRestAPITask(APIView):
    def post(self, request, notebook_slug):
        try:
            notebook = (
                notebooks_queryset(request).filter(slug=notebook_slug).latest("id")
            )
        except Notebook.DoesNotExist:
            raise Http404()
        try:
            with transaction.atomic():
                task = Task(
                    session_id=uuid.uuid4().hex,
                    state="CREATED",
                    notebook=notebook,
                    params=json.dumps(request.data),
                )
                task.save()
                job_params = {"db_id": task.id}
                transaction.on_commit(lambda: task_execute.delay(job_params))
            return Response({"id": task.session_id}, status=status.HTTP_201_CREATED)
        except Exception as e:
            raise APIException(str(e))


class GetRestAPITask(APIView):
    def get(self, requset, session_id):
        try:
            task = Task.objects.filter(
                session_id=session_id,
            ).latest("id")

            if task.state == "DONE":
                response = {}
                try:
                    fname = os.path.join(
                        settings.MEDIA_ROOT, task.session_id, "response.json"
                    )
                    if os.path.exists(fname):
                        with open(fname) as fin:
                            response = json.loads(fin.read())
                except Exception as e:
                    return Response(
                        str(e), status=status.HTTP_500_INTERNAL_SERVER_ERROR
                    )
                return Response(response, status=status.HTTP_200_OK)
            if task.state == "ERROR":
                return Response(
                    task.result, status=status.HTTP_500_INTERNAL_SERVER_ERROR
                )
            return Response({"state": "running"}, status=status.HTTP_202_ACCEPTED)
        except Task.DoesNotExist:
            raise Http404()


class ExportPDF(APIView):
    def post(self, request):
        try:
            # check if user can access the notebook
            notebook = notebooks_queryset(request).get(pk=request.data["notebook_id"])
        except Notebook.DoesNotExist:
            raise Http404()
        try:
            celery_job = export_to_pdf.delay(request.data)
            return Response({"job_id": celery_job.id}, status=status.HTTP_201_CREATED)
        except Exception as e:
            raise APIException(str(e))


class GetPDFAddress(APIView):
    def get(self, request, job_id):
        res = AsyncResult(job_id)
        fileUrl, title, error = "", "", ""
        if res.ready():
            if res.state == "FAILURE":
                error = str(res.result)
            elif res.state == "SUCCESS":
                fileUrl, title = res.result
        return Response(
            {"ready": res.ready(), "url": fileUrl, "title": title, "error": error}
        )


class ExecutionHistoryView(ListAPIView):
    serializer_class = TaskSerializer
    queryset = Task.objects.all()

    def get_queryset(self):
        # check if user has access to the notebook
        notebook = get_object_or_404(
            notebooks_queryset(self.request), pk=self.kwargs["notebook_id"]
        )

        return Task.objects.filter(
            notebook_id=notebook.id,
            session_id=self.kwargs["session_id"],
        )




def get_bucket_key(site, user, filename):
    return f"site-{site.id}/user-{user.id}/{filename}"


def get_site_bucket_key(site, filename):
    return f"site-{site.id}/files/{filename}"


def get_worker_bucket_key(session_id, output_dir, filename):
    return f"session-{session_id}/{output_dir}/{filename}"


def get_user_upload_bucket_key(site_id, session_id, filename):
    return f"site-{site_id}/session-{session_id}/user-input/{filename}"




from django.conf.urls import url

from apps.workers.views import (
    DeleteWorker,
    GetWorker,
    IsWorkerStale,
    SetWorkerState,
    WorkerGetNb,
    WorkerUpdateNb,
)

workers_urlpatterns = [
    url(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/nb",
        WorkerGetNb.as_view(),
    ),
    url(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/update-nb",
        WorkerUpdateNb.as_view(),
    ),
    url(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/worker",
        GetWorker.as_view(),
    ),
    url(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/is-worker-stale",
        IsWorkerStale.as_view(),
    ),
    url(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/set-worker-state",
        SetWorkerState.as_view(),
    ),
    url(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/delete-worker",
        DeleteWorker.as_view(),
    ),
]



# Generated by Django 3.2.5 on 2023-03-28 08:50

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("accounts", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Notebook",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("file_updated_at", models.DateTimeField(blank=True)),
                ("title", models.CharField(max_length=512)),
                ("slug", models.CharField(blank=True, max_length=512)),
                ("path", models.CharField(max_length=1024)),
                (
                    "image_path",
                    models.CharField(blank=True, default="", max_length=1024),
                ),
                ("params", models.TextField(blank=True)),
                ("state", models.CharField(blank=True, max_length=128)),
                ("default_view_path", models.CharField(blank=True, max_length=1024)),
                ("output", models.CharField(blank=True, max_length=128)),
                ("format", models.CharField(blank=True, max_length=1024)),
                ("schedule", models.CharField(blank=True, max_length=128)),
                ("notify", models.TextField(blank=True)),
                ("errors", models.TextField(blank=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "hosted_on",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="accounts.site"
                    ),
                ),
            ],
        ),
    ]



import json
import logging
import os
import sys
from types import SimpleNamespace

import requests

from apps.workers.models import Worker, WorkerState
from apps.ws.utils import machine_uuid

log = logging.getLogger(__name__)


class RESTClient:
    def __init__(self, notebook_id, session_id, worker_id):
        self.server_url = os.environ.get("MERCURY_SERVER_URL", "http://127.0.0.1:8000")
        self.notebook_id = notebook_id
        self.session_id = session_id
        self.worker_id = worker_id
        self.worker = None  # SimpleNamespace object
        self.state = WorkerState.Unknown
        self.notebook = None  # SimpleNamespace object
        self.load_notebook()

    def load_notebook(self):
        try:
            log.debug(f"Load notebook id={self.notebook_id}")
            response = requests.get(
                f"{self.server_url}/api/v1/worker/{self.session_id}/{self.worker_id}/{self.notebook_id}/nb"
            )
            if response.status_code != 200:
                raise Exception("Cant load notebook")
            self.notebook = SimpleNamespace(**response.json())
        except Exception:
            log.exception("Exception when notebook load, quit")
            sys.exit(0)

    def update_notebook(self, new_params):
        log.debug(f"Executed params {json.dumps(new_params, indent=4)}")

        update_database = False
        if new_params.get("title", "") != "" and self.notebook.title != new_params.get(
            "title", ""
        ):
            self.notebook.title = new_params.get("title", "")
            update_database = True

        nb_params = json.loads(self.notebook.params)
        for property in [
            "show-code",
            "show-prompt",
            "continuous_update",
            "static_notebook",
            "description",
            "show_sidebar",
            "full_screen",
            "allow_download",
        ]:
            if new_params.get(property) is not None and nb_params.get(
                property
            ) != new_params.get(property):
                nb_params[property] = new_params.get(property)
                update_database = True
        # save widgets params
        if json.dumps(nb_params.get("params", {})) != json.dumps(
            new_params.get("params", {})
        ):
            nb_params["params"] = new_params["params"]
            update_database = True

        if update_database:
            try:
                log.debug(f"Update notebook id={self.notebook_id}")
                response = requests.post(
                    f"{self.server_url}/api/v1/worker/{self.session_id}/{self.worker_id}/{self.notebook_id}/update-nb",
                    {"title": self.notebook.title, "params": json.dumps(nb_params)},
                )
                if response.status_code != 200:
                    raise Exception(f"Cant update notebook {response}")
                self.notebook = SimpleNamespace(**response.json())
            except Exception:
                log.exception("Exception when updating notebook")

        return update_database

    def is_presentation(self):
        try:
            isIt = self.notebook.output == "slides"
            log.debug(f"Check if notebook is presentation ({isIt})")
            return isIt
        except Exception:
            log.exception("Exception when check if notebook is presentation")
        return False

    def show_code(self):
        try:
            show_it = str(
                json.loads(self.notebook.params).get("show-code", "false")
            ).lower()
            log.debug(f"Check if show code from notebook ({show_it})")
            return show_it == "true"
        except Exception:
            log.exception("Exception when check if show code from notebook")
        return False

    def show_prompt(self):
        try:
            show_it = str(
                json.loads(self.notebook.params).get("show-prompt", "false")
            ).lower()
            log.debug(f"Check if show prompt from notebook ({show_it})")
            return show_it == "true"
        except Exception:
            log.exception("Exception when check if show promtp from notebook")
        return False

    def reveal_theme(self):
        # TODO: get reveal theme
        return "white"

    def worker_state(self):
        return self.state

    def set_worker_state(self, new_state):
        try:
            log.debug(
                f"Worker id={self.worker_id} set state {new_state} uuid {machine_uuid()}"
            )
            self.state = new_state
            # set worker machine id
            # to control number of workers
            # in the single machine
            response = requests.post(
                f"{self.server_url}/api/v1/worker/{self.session_id}/{self.worker_id}/{self.notebook_id}/set-worker-state",
                {"state": new_state, "machine_id": machine_uuid()},
            )
            if response.status_code != 200:
                raise Exception(f"Problem when set worker state {response}")
            self.worker = SimpleNamespace(**response.json())
        except Exception:
            log.exception("Exception when set worker state")

    @staticmethod
    def delete_worker_in_db(session_id, worker_id, notebook_id):
        try:
            log.debug(f"Delete worker id={worker_id}")
            server_url = os.environ.get("MERCURY_SERVER_URL", "http://127.0.0.1:8000")

            response = requests.post(
                f"{server_url}/api/v1/worker/{session_id}/{worker_id}/{notebook_id}/delete-worker",
            )
            if response.status_code != 204:
                raise Exception(f"Problem when delete worker {response}")

        except Exception:
            pass
            # log.exception(f"Exception when delete worker")

    def delete_worker(self):
        RESTClient.delete_worker_in_db(
            self.session_id, self.worker_id, self.notebook_id
        )

    def worker_exists(self):
        try:
            log.debug(f"Worker id={self.worker_id} exists")

            response = requests.get(
                f"{self.server_url}/api/v1/worker/{self.session_id}/{self.worker_id}/{self.notebook_id}/worker",
            )
            self.worker = SimpleNamespace(**response.json())

        except Exception as e:
            # log.exception(f"Worker id={self.worker_id} does not exists, quit")
            sys.exit(1)
        return True

    def is_worker_stale(self):
        try:
            response = requests.get(
                f"{self.server_url}/api/v1/worker/{self.session_id}/{self.worker_id}/{self.notebook_id}/is-worker-stale",
            )
            if response.status_code == 200:
                is_stale = response.json().get("is_stale", True)
                log.debug(f"Check worker id={self.worker_id} is stale {is_stale}")
                return is_stale
            return True

        except Exception:
            log.exception(
                f"Exception when check if worker id={self.worker_id} is stale"
            )
        return True



import os
import uuid

from django.template.defaultfilters import slugify


def is_cloud_version():
    return os.environ.get("MERCURY_CLOUD", "0") == "1"


def some_random_slug():
    h = uuid.uuid4().hex.replace("-", "")
    return h[:8]


def get_slug(slug, title):
    new_slug = slugify(slug)
    if new_slug == "":
        new_slug = slugify(title)
    if new_slug == "":
        new_slug = some_random_slug()
    return new_slug



from django.contrib.auth.models import User
from django.db import transaction
from rest_framework import permissions, viewsets
from rest_framework.exceptions import APIException

from apps.accounts.models import Membership, Site
from apps.accounts.serializers import MembershipSerializer

from apps.accounts.views.permissions import HasEditRights


import os
import json
import requests
from rest_framework import permissions, status
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.serializers import ValidationError

from apps.accounts.serializers import UserSerializer


class MembershipViewSet(viewsets.ModelViewSet):
    serializer_class = MembershipSerializer
    permission_classes = [permissions.IsAuthenticated, HasEditRights]

    def get_queryset(self):
        return Membership.objects.filter(host__id=self.kwargs["site_id"])

    def perform_create(self, serializer):
        try:
            # create a database instance
            with transaction.atomic():
                site = Site.objects.get(pk=self.kwargs.get("site_id"))
                user = User.objects.get(pk=self.request.data.get("user_id"))
                instance = serializer.save(
                    host=site, user=user, created_by=self.request.user
                )
                instance.save()
        except Exception as e:
            raise APIException(str(e))


class DeleteAccount(APIView):
    permission_classes = [permissions.IsAuthenticated]

    def post(self, request, format=None):
        user = request.user

        info = json.loads(user.profile.info)
        plan = info.get("plan", "starter")
        if plan != "starter":
            return Response(
                data={"msg": "Please cancel subscription"},
                status=status.HTTP_403_FORBIDDEN,
            )

        # remove other stuff ...
        print("TODO: remove all files ...")

        user.delete()

        return Response(status=status.HTTP_204_NO_CONTENT)



import logging
import os
import shutil
import sys
import uuid

import requests
from django.conf import settings
from execnb.nbio import write_nb
from django_drf_filepond.models import TemporaryUpload

from apps.nbworker.utils import stop_event
from apps.storage.models import UploadedFile
from apps.storage.utils import get_worker_bucket_key
from apps.tasks.export_pdf import to_pdf

log = logging.getLogger(__name__)


class StorageManager:
    def __init__(self, session_id, worker_id, notebook_id):
        self.session_id = session_id
        self.worker_id = worker_id
        self.notebook_id = notebook_id
        self.server_url = os.environ.get("MERCURY_SERVER_URL", "http://127.0.0.1:8000")
        if settings.STORAGE not in [settings.STORAGE_MEDIA, settings.STORAGE_S3]:
            log.error(f"{settings.STORAGE} not implemented")
            stop_event.set()
            sys.exit(1)

    @staticmethod
    def download_file(url):
        local_filename = url.split("?")[0].split("/")[-1]
        with requests.get(url, stream=True) as r:
            with open(local_filename, "wb") as f:
                shutil.copyfileobj(r.raw, f)

        return local_filename

    def provision_uploaded_files(self):
        log.debug(f"Provision uploaded files")
        if settings.STORAGE == settings.STORAGE_MEDIA:
            pass
        elif settings.STORAGE == settings.STORAGE_S3:
            # get links
            url = f"{self.server_url}/api/v1/worker/uploaded-files-urls/{self.session_id}/{self.worker_id}/{self.notebook_id}"
            response = requests.get(url)
            download_urls = response.json().get("urls")
            # download files
            for url in download_urls:
                f = StorageManager.download_file(url)
                log.debug(f"Downloaded {f}")

    @staticmethod
    def create_dir(dir_path):
        if not os.path.exists(dir_path):
            try:
                log.debug(f"Create directory {dir_path}")
                os.mkdir(dir_path)
            except Exception as e:
                log.exception(f"Exception when create {dir_path}")
                raise Exception(f"Cant create {dir_path}")
        else:
            log.debug(f"Directory {dir_path} already exists")

    @staticmethod
    def delete_dir(dir_path):
        if os.path.exists(dir_path):
            try:
                log.debug(f"Delete directory {dir_path}")
                shutil.rmtree(dir_path, ignore_errors=True)
            except Exception as e:
                log.exception(f"Exception when delete {dir_path}")
                raise Exception(f"Cant delete {dir_path}")

    def worker_output_dir(self):
        if settings.STORAGE == settings.STORAGE_MEDIA:
            first_dir = os.path.join(str(settings.MEDIA_ROOT), self.session_id)
            StorageManager.create_dir(first_dir)
            output_dir = os.path.join(
                str(settings.MEDIA_ROOT), self.session_id, f"output_{self.worker_id}"
            )
            StorageManager.create_dir(output_dir)
            log.debug(f"Worker output directory: {output_dir}")
            return output_dir
        elif settings.STORAGE == settings.STORAGE_S3:
            output_dir = f"output_{self.worker_id}"
            StorageManager.create_dir(output_dir)
            return output_dir

    def delete_worker_output_dir(self):
        if settings.STORAGE == settings.STORAGE_MEDIA:
            first_dir = os.path.join(str(settings.MEDIA_ROOT), self.session_id)
            log.debug(f"Delete Worker output directory: {first_dir}")
            StorageManager.delete_dir(first_dir)
        elif settings.STORAGE == settings.STORAGE_S3:
            output_dir = f"output_{self.worker_id}"
            StorageManager.delete_dir(output_dir)

    def sync_output_dir(self):
        if settings.STORAGE == settings.STORAGE_MEDIA:
            # nothing to do
            pass
        elif settings.STORAGE == settings.STORAGE_S3:
            # list all files in directory
            local_files = []
            output_dir = self.worker_output_dir()
            with os.scandir(output_dir) as entries:
                for entry in entries:
                    if entry.is_file():
                        print(entry, entry.path)
                        local_files += [entry]
            log.debug(f"Sync {local_files}")

            action = "put_object"
            for entry in local_files:
                # upload them to s3
                url = f"{self.server_url}/api/v1/worker/presigned-url/{action}/{self.session_id}/{self.worker_id}/{self.notebook_id}/{output_dir}/{entry.name}"
                response = requests.get(url)
                upload_url = response.json().get("url")

                with open(entry.path, "rb") as fin:
                    response = requests.put(upload_url, fin.read())

                # create db objects
                url = f"{self.server_url}/api/v1/worker/add-file"
                data = {
                    "worker_id": self.worker_id,
                    "session_id": self.session_id,
                    "notebook_id": self.notebook_id,
                    "filename": entry.name,
                    "filepath": get_worker_bucket_key(
                        self.session_id, output_dir, entry.name
                    ),
                    "output_dir": output_dir,
                    "local_filepath": entry.path,
                }
                response = requests.post(url, data)
                # that's all

    def list_worker_files_urls(self):
        files_urls = []
        if settings.STORAGE == settings.STORAGE_MEDIA:
            output_dir = self.worker_output_dir()
            for f in os.listdir(output_dir):
                if os.path.isfile(os.path.join(output_dir, f)):
                    files_urls += [
                        f"{settings.MEDIA_URL}/{self.session_id}/output_{self.worker_id}/{f}"
                    ]
        elif settings.STORAGE == settings.STORAGE_S3:
            pass

        return files_urls

    # def save_nb(self, nb):
    #     fpath = None
    #     if settings.STORAGE == settings.STORAGE_MEDIA:
    #         fpath = os.path.join(
    #             self.worker_output_dir(), f"nb-{self.some_hash()}.ipynb"
    #         )
    #         write_nb(nb, fpath)
    #     return fpath

    def save_nb_html(self, nb_html_body):
        html_path, url = None, None
        fname = f"download-notebook-{self.some_hash()}.html"

        if settings.STORAGE == settings.STORAGE_MEDIA:
            html_path = os.path.join(self.worker_output_dir(), fname)
            with open(html_path, "w", encoding="utf-8", errors="ignore") as fout:
                fout.write(nb_html_body)
            html_url = f"{settings.MEDIA_URL}/{self.session_id}/output_{self.worker_id}/{fname}"
        elif settings.STORAGE == settings.STORAGE_S3:
            # 1.
            # get upload link
            action = "put_object"
            output_dir = "downdload-html"
            url = f"{self.server_url}/api/v1/worker/presigned-url/{action}/{self.session_id}/{self.worker_id}/{self.notebook_id}/{output_dir}/{fname}"
            response = requests.get(url)
            upload_url = response.json().get("url")
            # 2.
            # upload file
            response = requests.put(upload_url, nb_html_body.encode("utf-8"))
            if response.status_code != 200:
                raise Exception(f"Notebook not uploaded {response}")
            # 3.
            # get download link
            action = "get_object"
            output_dir = "downdload-html"
            url = f"{self.server_url}/api/v1/worker/presigned-url/{action}/{self.session_id}/{self.worker_id}/{self.notebook_id}/{output_dir}/{fname}"
            response = requests.get(url)
            html_url = response.json().get("url")

        return html_path, html_url

    def save_nb_pdf(self, nb_html_body, is_presentation):
        pdf_path, pdf_url = None, None
        fname = f"download-notebook-{self.some_hash()}.pdf"
        if settings.STORAGE == settings.STORAGE_MEDIA:
            # save HTML
            html_path, html_url = self.save_nb_html(nb_html_body)

            # check if we need postfix
            slides_postfix = "?print-pdf" if is_presentation else ""

            # export to PDF
            pdf_path = html_path.replace(".html", ".pdf")
            pdf_url = html_url.replace(".html", ".pdf")
            log.debug(f"Export {html_path}{slides_postfix} to PDF {pdf_path}")
            to_pdf(f"{html_path}{slides_postfix}", pdf_path)

        elif settings.STORAGE == settings.STORAGE_S3:
            # 0. first lets create HTML file
            html_fname = fname.replace(".pdf", ".html")
            html_path = os.path.abspath(
                os.path.join(self.worker_output_dir(), html_fname)
            )
            pdf_path = os.path.abspath(os.path.join(self.worker_output_dir(), fname))

            log.debug(f"Dump to HTML {html_path}")
            with open(html_path, "w", encoding="utf-8", errors="ignore") as fout:
                fout.write(nb_html_body)
            # check if we need postfix
            slides_postfix = "?print-pdf" if is_presentation else ""
            # export to PDF
            log.debug(f"Export HTML {html_path}{slides_postfix} to PDF {pdf_path}")
            to_pdf(f"{html_path}{slides_postfix}", pdf_path)

            # 1.
            # get upload link
            action = "put_object"
            output_dir = "downdload-pdf"
            url = f"{self.server_url}/api/v1/worker/presigned-url/{action}/{self.session_id}/{self.worker_id}/{self.notebook_id}/{output_dir}/{fname}"
            response = requests.get(url)
            upload_url = response.json().get("url")
            # 2.
            # upload file
            with open(pdf_path, "rb") as fout:
                response = requests.put(upload_url, fout.read())
                if response.status_code != 200:
                    raise Exception(f"PDF notebook not uploaded {response}")
            # 3.
            # get download link
            action = "get_object"
            output_dir = "downdload-pdf"
            url = f"{self.server_url}/api/v1/worker/presigned-url/{action}/{self.session_id}/{self.worker_id}/{self.notebook_id}/{output_dir}/{fname}"
            response = requests.get(url)
            pdf_url = response.json().get("url")

        return pdf_path, pdf_url

    def some_hash(self):
        h = uuid.uuid4().hex.replace("-", "")
        return h[:8]


    def get_user_uploaded_file(self, value):
        if settings.STORAGE == settings.STORAGE_MEDIA:
            log.debug(f"Get file {value[0]} from id={value[1]}")
            tu = TemporaryUpload.objects.get(upload_id=value[1])
            value[1] = tu.get_file_path()
            value[1] = value[1].replace("\\", "\\\\")
            log.debug(f"File path is {value[1]}")
        elif settings.STORAGE == settings.STORAGE_S3:
            # get link
            
            url = f"{self.server_url}/api/v1/worker/user-uploaded-file/{self.session_id}/{self.worker_id}/{self.notebook_id}/{value[0]}"
            print(url)
            response = requests.get(url)
            download_url = response.json().get("url")
            # download file
            f = StorageManager.download_file(download_url)
            log.debug(f"Downloaded {f}")
            value[1] = f
        return value

    




from channels.db import database_sync_to_async
from channels.middleware import BaseMiddleware
from django.contrib.auth.models import AnonymousUser
from rest_framework.authtoken.models import Token


@database_sync_to_async
def get_user(token_key):
    try:
        token = Token.objects.get(key=token_key)
        return token.user
    except Token.DoesNotExist:
        return AnonymousUser()


class TokenAuthMiddleware(BaseMiddleware):
    def __init__(self, inner):
        super().__init__(inner)

    async def __call__(self, scope, receive, send):
        try:
            token_key = (
                dict((x.split("=") for x in scope["query_string"].decode().split("&")))
            ).get("token", None)
        except ValueError:
            token_key = None
        scope["user"] = (
            AnonymousUser() if token_key is None else await get_user(token_key)
        )
        return await super().__call__(scope, receive, send)



from allauth.account.admin import EmailAddress
from celery import shared_task

from django.conf import settings
from django.core.mail import send_mail

from apps.accounts.models import Invitation, Membership, Site, SiteStatus
from apps.notebooks.models import Notebook
from apps.notebooks.tasks import task_init_notebook
from apps.storage.models import UploadedFile
from apps.storage.s3utils import S3
from apps.storage.utils import get_site_bucket_key


@shared_task(bind=True)
def task_init_site(self, job_params):
    site = Site.objects.get(pk=job_params["site_id"])
    site.status = SiteStatus.INITIALIZING
    site.save()

    files = UploadedFile.objects.filter(hosted_on=site)

    any_notebooks = len([f for f in files if f.filetype == "ipynb"]) > 0

    if any_notebooks:
        # clear previous notebooks
        # just delete them
        Notebook.objects.filter(hosted_on=site).delete()
        s3 = S3()
        for f in files:
            print(f"Download {f.filepath}")
            s3.download_file(f.filepath, f.filename)
            if f.filetype == "ipynb":
                task_init_notebook(
                    f.filename,
                    bucket_key=get_site_bucket_key(site, "<replace>"),
                    site=site,
                    user=f.created_by,
                )

    site.status = SiteStatus.READY
    site.save()


def get_app_address(site):
    subdomain = site.slug
    domain = site.domain
    custom_domain = site.custom_domain

    if custom_domain is not None and custom_domain != "":
        return custom_domain

    return f"https://{subdomain}.{domain}"


@shared_task(bind=True)
def task_send_invitation(self, job_params):
    invitation_id = job_params["invitation_id"]
    invitation = Invitation.objects.get(pk=invitation_id)

    from_address = EmailAddress.objects.get(user=invitation.created_by, primary=True)
    invited_by = invitation.created_by

    send_mail(
        "Mercury Invitation",
        f"""Hi,

User {invited_by.username} invites you to {invitation.rights.lower()} web app at {get_app_address(invitation.hosted_on)}.

Please create a new account at https://cloud.runmercury.com to access app.

Thank you!
Mercury Team        
""",
        settings.DEFAULT_FROM_EMAIL,
        [invitation.invited],
        fail_silently=False,
    )


@shared_task(bind=True)
def task_send_new_member(self, job_params):
    membership_id = job_params["membership_id"]
    membership = Membership.objects.get(pk=membership_id)

    from_address = EmailAddress.objects.get(user=membership.created_by, primary=True)
    invited_by = membership.created_by

    send_mail(
        "Mercury Invitation",
        f"""Hi,

User {invited_by.username} invites you to {membership.rights.lower()} web app at {get_app_address(membership.host)}.

Thank you!
Mercury Team        
""",
        settings.DEFAULT_FROM_EMAIL,
        [membership.user.email],
        fail_silently=False,
    )



import json
import os
import uuid
from datetime import datetime, timedelta

from django.contrib.auth.models import User
from django.db import transaction
from django.db.models import Q
from django.template.defaultfilters import slugify
from rest_framework import permissions, status, viewsets
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.accounts.models import Invitation, Membership, Site
from apps.accounts.serializers import InvitationSerializer
from apps.accounts.tasks import (
    task_init_site,
    task_send_invitation,
    task_send_new_member,
)
from apps.accounts.views.permissions import HasEditRights


class InviteView(APIView):
    permission_classes = [permissions.IsAuthenticated, HasEditRights]

    def post(self, request, site_id, format=None):
        try:
            # create a database instance
            with transaction.atomic():
                address_email = request.data.get("email")
                rights = request.data.get("rights", "VIEW")
                if rights not in ["VIEW", "EDIT"]:
                    rights = "VIEW"

                site = Site.objects.get(pk=site_id)

                # check for users
                already_user = User.objects.filter(email=address_email)

                if already_user:
                    membership = Membership.objects.create(
                        user=already_user[0],
                        host=site,
                        rights=rights,
                        created_by=request.user,
                    )
                    job_params = {"membership_id": membership.id}
                    transaction.on_commit(
                        lambda: task_send_new_member.delay(job_params)
                    )

                else:
                    invitation = Invitation.objects.create(
                        invited=address_email,
                        created_by=request.user,
                        rights=rights,
                        hosted_on=site,
                    )

                    job_params = {"invitation_id": invitation.id}
                    transaction.on_commit(
                        lambda: task_send_invitation.delay(job_params)
                    )

                return Response(status=status.HTTP_200_OK)
        except Exception as e:
            raise APIException(str(e))


class ListInvitations(APIView):
    permission_classes = [permissions.IsAuthenticated, HasEditRights]

    def get(self, request, site_id, format=None):
        try:
            invitations = Invitation.objects.filter(hosted_on__id=site_id)
            return Response(
                InvitationSerializer(invitations, many=True).data,
                status=status.HTTP_200_OK,
            )
        except Exception as e:
            print(str(e))
            raise APIException(str(e))


class DeleteInvitation(APIView):
    permission_classes = [permissions.IsAuthenticated, HasEditRights]

    def delete(self, request, site_id, invitation_id, format=None):
        try:
            invitation = Invitation.objects.get(pk=invitation_id, hosted_on__id=site_id)
            invitation.delete()
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception as e:
            raise APIException(str(e))



import os

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator
from django.core.asgi import get_asgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "server.settings")
# Initialize Django ASGI application early to ensure the AppRegistry
# is populated before importing code that may import ORM models.
django_asgi_app = get_asgi_application()

import apps.ws.routing
from apps.ws.middleware import TokenAuthMiddleware

application = ProtocolTypeRouter(
    {
        "http": django_asgi_app,
        "websocket": TokenAuthMiddleware(  # AllowedHostsOriginValidator(
            AuthMiddlewareStack(URLRouter(apps.ws.routing.websocket_urlpatterns))
        )
        # ),
    }
)



import os

from django.conf import settings
from django.contrib.auth.models import User
from django.core.mail import EmailMessage

from apps.notebooks.models import Notebook
from apps.tasks.tasks_export import export_to_pdf


def validate_notify(config: dict) -> str:
    """returns the string with error message from notify parsing"""

    if not config:
        return ""
    try:
        on_success, on_failure, attachment = parse_config(config)
        if not on_success and not on_failure:
            return (
                "Please specify `on_success` or `on_failure` in the `notify` parameter."
            )
        if attachment and not ("pdf" in attachment or "html" in attachment):
            return "Please specify `html` or `pdf` format in the `attachment` in the `notify`."
    except Exception as e:
        return "Error while parsing `notify` in the YAML config."
    return ""


def username_to_email(username):
    users = User.objects.filter(username=username)
    if not users:
        return ""
    # not worry about duplicates in username
    return users[0].email


def list_to_emails(contacts):
    emails = []
    for contact in contacts:
        contact = contact.strip()
        if "@" in contact:
            emails += [contact]
        else:
            email_address = username_to_email(contact)
            if email_address:
                emails += [email_address]
    return list(set(emails))


def parse_config(config):
    on_success = config.get("on_success", "")
    on_failure = config.get("on_failure", "")
    attachment = config.get("attachment", "")

    on_success = on_success.split(",")
    on_failure = on_failure.split(",")

    on_success = list_to_emails(on_success)
    on_failure = list_to_emails(on_failure)

    return on_success, on_failure, attachment


def notify(config, is_success, error_msg, notebook_id, notebook_url):
    try:
        if not config:
            # no config provided, skip notify step
            return

        on_success, on_failure, attachment = parse_config(config)

        notebook = Notebook.objects.get(pk=notebook_id)

        email = None
        if on_success and is_success:
            msg = f"""Notebook '{notebook.title}' executed successfully."""
            email = EmailMessage(
                "Notebook executed successfully",
                msg,
                None,  # use default from email (from django settings)
                on_success,
            )
        if on_failure and not is_success:
            msg = f"""Notebook '{notebook.title}' failed to execute. {error_msg}"""
            email = EmailMessage(
                "Notebook failed to execute",
                msg,
                None,  # use default from email (from django settings)
                on_failure,
            )
        if email is not None:
            notebook_html_path = os.path.join(
                *(
                    [settings.MEDIA_ROOT]
                    + notebook_url.replace(settings.MEDIA_URL, "", 1).split("/")
                )
            )

            if "html" in attachment and os.path.exists(notebook_html_path):
                email.attach_file(notebook_html_path)
            if "pdf" in attachment and os.path.exists(notebook_html_path):
                export_to_pdf(
                    {"notebook_id": notebook_id, "notebook_path": notebook_url}
                )
                notebook_pdf_path = notebook_html_path.replace(".html", ".pdf")
                email.attach_file(notebook_pdf_path)

            email.send(fail_silently=True)
    except Exception as e:
        print("Error in the notify step")
        print(str(e))



from django.contrib.auth.models import User
from django.db import models

from apps.accounts.models import Site


class Notebook(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    file_updated_at = models.DateTimeField(blank=True)
    title = models.CharField(max_length=512, blank=False)
    slug = models.CharField(max_length=512, blank=True)
    path = models.CharField(max_length=1024, blank=False)
    image_path = models.CharField(default="", max_length=1024, blank=True)
    params = models.TextField(blank=True)
    state = models.CharField(max_length=128, blank=True)
    default_view_path = models.CharField(max_length=1024, blank=True)
    output = models.CharField(max_length=128, blank=True)
    format = models.CharField(max_length=1024, blank=True)
    schedule = models.CharField(max_length=128, blank=True)
    notify = models.TextField(blank=True)
    errors = models.TextField(blank=True)

    created_by = models.ForeignKey(User, on_delete=models.CASCADE)
    hosted_on = models.ForeignKey(Site, on_delete=models.CASCADE)



import json
import logging
import time
import sys
from queue import Queue

import websocket

from apps.nbworker.rest import RESTClient
from apps.nbworker.utils import Purpose, stop_event
from apps.workers.models import WorkerState
from apps.storage.storage import StorageManager

log = logging.getLogger(__name__)


class WSClient(RESTClient):
    def __init__(self, ws_address, notebook_id, session_id, worker_id):
        super(WSClient, self).__init__(notebook_id, session_id, worker_id)

        self.sm = StorageManager(self.session_id, self.worker_id, self.notebook_id)

        self.ws_address = ws_address

        self.connect(ws_address)

        self.queue = Queue()

        self.msg_counter = 0

    def connect(self, ws_address):
        try:
            log.debug(f"WS connect to {ws_address}")
            self.ws = websocket.WebSocketApp(
                ws_address,
                on_open=lambda ws: self.on_open(ws),
                on_close=lambda ws, close_status_code, close_msg: self.on_close(
                    ws, close_status_code, close_msg
                ),
                on_error=lambda ws, msg: self.on_error(ws, msg),
                on_pong=lambda ws, msg: self.on_pong(ws, msg),
                on_message=lambda ws, msg: self.on_message(ws, msg),
            )
        except Exception:
            log.exception("Exception when WS connect")

    def on_open(self, ws):
        log.info("Open ws connection")
        self.queue.put(json.dumps({"purpose": Purpose.InitNotebook}))
        if self.worker_exists():
            self.set_worker_state(WorkerState.Running)
            self.send_state()

    def on_close(self, ws, close_status_code, close_msg):
        global stop_event
        stop_event.set()
        self.sm.delete_worker_output_dir()
        log.info(f"WS close connection, status={close_status_code}, msg={close_msg}")

    def on_pong(self, wsapp, msg):
        log.info("WS on_pong")
        if self.is_worker_stale():
            self.delete_worker()
            log.info(f"Worker id={self.worker_id} is stale, quit")
            sys.exit(1)

    def on_error(self, ws, msg):
        log.debug(f"WS on_error, {msg}")

    def on_message(self, ws, msg):
        log.debug(f"WS on_message {msg}")

        json_data = json.loads(msg)    
        if json_data.get("purpose", "") == Purpose.WorkerPing:
            self.worker_pong()
        else:
            self.queue.put(msg)

        self.msg_counter += 1

    def send_state(self):
        try:
            log.debug(f"Send state {self.worker_state()}")
            msg = {
                "purpose": Purpose.WorkerState,
                "state": self.worker_state(),
                "workerId": self.worker_id,
            }
            self.ws.send(json.dumps(msg))
        except Exception as e:
            log.exception("Exception when send state")

    def update_worker_state(self, new_state):
        self.set_worker_state(new_state)
        self.send_state()



# Generated by Django 3.2.5 on 2023-03-28 08:50

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("notebooks", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Task",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("task_id", models.CharField(blank=True, max_length=128)),
                ("session_id", models.CharField(max_length=128)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("state", models.CharField(blank=True, max_length=128)),
                ("params", models.TextField(blank=True)),
                ("result", models.TextField(blank=True)),
                (
                    "notebook",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="notebooks.notebook",
                    ),
                ),
            ],
        ),
    ]



from django.urls import re_path

from apps.storage.views.dashboardfiles import (
    GetStorageType,
    DeleteFile,
    FileUploaded,
    ListFiles,
    PresignedUrl,
    PresignedUrlPut,
)
from apps.storage.views.workerfiles import (
    WorkerAddFile,
    WorkerGetUploadedFilesUrls,
    WorkerPresignedUrl,
)
from apps.storage.views.notebookfiles import (
    NbPresignedUrlPut,
    NbFileUploaded,
    NbDeleteFile,
    WorkerGetNbFileUrl,
)

storage_urlpatterns = [
    re_path(
        "api/v1/storage-type",
        GetStorageType.as_view(),
    ),
    #
    # dashboard files
    #
    re_path("api/v1/(?P<site_id>.+)/files", ListFiles.as_view()),
    re_path(
        "api/v1/presigned-url/(?P<action>.+)/(?P<site_id>.+)/(?P<filename>.+)",
        PresignedUrl.as_view(),
    ),
    re_path(
        "api/v1/presigned-url-put/(?P<site_id>.+)/(?P<filename>.+)/(?P<filesize>.+)",
        PresignedUrlPut.as_view(),
    ),
    re_path(
        "api/v1/file-uploaded",
        FileUploaded.as_view(),
    ),
    re_path(
        "api/v1/delete-file",
        DeleteFile.as_view(),
    ),
    #
    # worker files
    #
    re_path(
        "api/v1/worker/presigned-url/(?P<action>.+)/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/(?P<output_dir>.+)/(?P<filename>.+)",
        WorkerPresignedUrl.as_view(),
    ),
    re_path(
        "api/v1/worker/add-file",
        WorkerAddFile.as_view(),
    ),
    re_path(
        "api/v1/worker/uploaded-files-urls/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)",
        WorkerGetUploadedFilesUrls.as_view(),
    ),
    # user uploaded files in notebooks
    re_path(
        "api/v1/nb-file-put/(?P<site_id>.+)/(?P<session_id>.+)/(?P<filename>.+)/(?P<filesize>.+)",
        NbPresignedUrlPut.as_view(),
    ),
    re_path(
        "api/v1/nb-file-uploaded",
        NbFileUploaded.as_view(),
    ),
    re_path(
        "api/v1/nb-delete-file",
        NbDeleteFile.as_view(),
    ),
    re_path(
        "api/v1/worker/user-uploaded-file/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/(?P<filename>.+)",
        WorkerGetNbFileUrl.as_view(),
    ),
]



import json
from enum import Enum

from django.contrib.auth.models import User
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver

from apps.accounts.fields import AutoCreatedField, AutoLastModifiedField
from apps.accounts.views.utils import is_cloud_version


class SiteStatus(str, Enum):
    CREATED = "Created"
    INITIALIZING = "Initializing"
    READY = "Ready"
    ERROR = "Error"


class Site(models.Model):
    title = models.CharField(
        max_length=256, help_text="Name of Mercury Site", blank=False, null=False
    )
    slug = models.CharField(
        max_length=256,
        help_text="Subdomain",
        blank=False,
        null=False,
        unique=True,
    )
    domain = models.CharField(
        default="runmercury.com",
        max_length=256,
        help_text="Domain address",
        blank=True,
        null=True,
    )
    custom_domain = models.CharField(
        max_length=256,
        help_text="Custom domain address",
        blank=True,
        null=True,
    )

    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    SHARE_CHOICES = (
        (PUBLIC, "Anyone can access notebooks and execute"),
        (PRIVATE, "Only selected users have access to notebooks"),
    )
    share = models.CharField(
        default=PUBLIC, max_length=32, choices=SHARE_CHOICES, blank=False, null=False
    )

    # Created
    # Initializing
    # Ready
    # Error
    status = models.CharField(default="Ready", max_length=32, blank=False, null=False)
    welcome = models.TextField(default="", blank=True, null=True)
    info = models.TextField(blank=True, null=True)

    active = models.BooleanField(default=True, blank=False, null=False)
    created_at = AutoCreatedField()
    updated_at = AutoLastModifiedField()
    created_by = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )


class UserProfile(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE, related_name="profile")
    # custom fields for user
    info = models.TextField(default="{}", blank=True)
    usage = models.TextField(default="{}", blank=True)


@receiver(post_save, sender=User)
def create_user_profile(sender, instance, created, **kwargs):
    if created:
        info = "{}"
        if is_cloud_version():
            info = json.dumps({"plan": "starter"})
        UserProfile.objects.create(user=instance, info=info)


@receiver(post_save, sender=User)
def save_user_profile(sender, instance, **kwargs):
    instance.profile.save()


@receiver(post_save, sender=User)
def check_invitations(sender, instance, **kwargs):
    invitations = Invitation.objects.filter(invited=instance.email)
    for invitation in invitations:
        previous_memberships = Membership.objects.filter(
            user=instance, host=invitation.hosted_on
        )
        if not previous_memberships:
            Membership.objects.create(
                user=instance,
                host=invitation.hosted_on,
                rights=invitation.rights,
                created_by=invitation.created_by,
            )
            invitation.delete()


class Membership(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    host = models.ForeignKey(Site, on_delete=models.CASCADE, related_name="hosts")
    # view, edit, admin
    VIEW = "VIEW"
    EDIT = "EDIT"
    RIGHTS_CHOICES = (
        (VIEW, "View and execute notebooks"),
        (EDIT, "Edit and view site, files and execute notebooks"),
    )
    rights = models.CharField(
        default=VIEW,
        choices=RIGHTS_CHOICES,
        max_length=32,
        help_text="Rights for user",
        blank=False,
        null=False,
    )
    created_at = AutoCreatedField()
    updated_at = AutoLastModifiedField()
    created_by = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="created_by"
    )


class Invitation(models.Model):
    invited = models.CharField(max_length=256, blank=False, null=False)
    created_at = AutoCreatedField()
    created_by = models.ForeignKey(User, on_delete=models.CASCADE)
    hosted_on = models.ForeignKey(Site, on_delete=models.CASCADE)
    rights = models.CharField(
        default=Membership.VIEW,
        choices=Membership.RIGHTS_CHOICES,
        max_length=32,
        help_text="Rights for user",
        blank=False,
        null=False,
    )


class Secret(models.Model):
    name = models.CharField(max_length=256, blank=False, null=False)
    token = models.TextField(blank=False, null=False)
    created_at = AutoCreatedField()
    created_by = models.ForeignKey(User, on_delete=models.CASCADE)
    hosted_on = models.ForeignKey(Site, on_delete=models.CASCADE)



from django.conf.urls import include, url
from django.urls import path, re_path
from django.views.generic.base import TemplateView
from rest_framework.routers import DefaultRouter

from apps.accounts.views.sites import GetSiteView, InitializeSite, SiteViewSet

from apps.accounts.views.accounts import MembershipViewSet, DeleteAccount
from apps.accounts.views.invitations import (
    DeleteInvitation,
    InviteView,
    ListInvitations,
)
from apps.accounts.views.secrets import (
    AddSecret,
    DeleteSecret,
    ListSecrets,
    WorkerListSecrets,
)
from apps.accounts.views.subscription import SubscriptionView

router = DefaultRouter()
router.register(r"api/v1/sites", SiteViewSet, basename="sites")
router.register(r"api/v1/(?P<site_id>.+)/members", MembershipViewSet, basename="sites")

accounts_urlpatterns = router.urls

accounts_urlpatterns += [
    path("api/v1/auth/", include("dj_rest_auth.urls")),
    path("api/v1/auth/register/", include("dj_rest_auth.registration.urls")),
    # path to set verify email in the frontend
    # fronted will do POST request to server with key
    # this is empty view, just to make reverse works
    re_path(
        r"^verify-email/(?P<key>[-:\w]+)/$",
        TemplateView.as_view(),
        name="account_confirm_email",
    ),
    # path to set password reset in the frontend
    # fronted will do POST request to server with uid and token
    # this is empty view, just to make reverse works
    re_path(
        r"^reset-password/(?P<uid>[-:\w]+)/(?P<token>[-:\w]+)/$",
        TemplateView.as_view(),
        name="password_reset_confirm",
    ),
    # sites
    re_path("api/v1/get-site/(?P<site_slug>.+)/", GetSiteView.as_view()),
    re_path("api/v1/init-site/(?P<site_id>.+)/", InitializeSite.as_view()),
    # invitations
    re_path("api/v1/(?P<site_id>.+)/invite", InviteView.as_view()),
    re_path("api/v1/(?P<site_id>.+)/list-invitations", ListInvitations.as_view()),
    re_path(
        "api/v1/(?P<site_id>.+)/delete-invitation/(?P<invitation_id>.+)",
        DeleteInvitation.as_view(),
    ),
    # secrets
    re_path("api/v1/(?P<site_id>.+)/add-secret", AddSecret.as_view()),
    re_path("api/v1/(?P<site_id>.+)/list-secrets", ListSecrets.as_view()),
    re_path(
        "api/v1/(?P<site_id>.+)/delete-secret/(?P<secret_id>.+)", DeleteSecret.as_view()
    ),
    re_path(
        "api/v1/worker/(?P<session_id>.+)/(?P<worker_id>.+)/(?P<notebook_id>.+)/worker-secrets",
        WorkerListSecrets.as_view(),
    ),
    re_path(
        "api/v1/subscription",
        SubscriptionView.as_view(),
    ),
    re_path("api/v1/auth/delete-account/", DeleteAccount.as_view()),
]



from rest_framework import serializers

from apps.storage.models import UploadedFile


class UploadedFileSerializer(serializers.ModelSerializer):
    class Meta:
        model = UploadedFile
        read_only_fields = ("id", "created_at", "created_by")
        fields = ("id", "created_at", "created_by", "filename", "filesize", "filetype")



import json
import logging
import os
import subprocess
import sys
import traceback
import uuid
from datetime import datetime
from shutil import which
from subprocess import PIPE, Popen

import nbformat
import yaml
from allauth.account.admin import EmailAddress
from celery import shared_task
from croniter import croniter
from django.conf import settings
from django.contrib.auth.models import User
from django.template.defaultfilters import slugify
from django.utils.timezone import make_aware

from apps.accounts.models import Site, SiteStatus
from apps.nb.exporter import Exporter
from apps.notebooks.models import Notebook
from apps.storage.s3utils import S3
from apps.tasks.models import Task
from apps.tasks.notify import validate_notify
from apps.ws.utils import parse_params

log = logging.getLogger(__name__)


def process_nbconvert_errors(error_msg):
    known_warnings = [
        "warn(",
        "UserWarning",
        "FutureWarning",
        "[NbConvertApp] Converting notebook",
        "[NbConvertApp] Writing",
    ]
    error_lines = []
    for e in error_msg.decode("utf-8").split("\n"):
        if e == "":
            continue
        known_warning = False
        for w in known_warnings:
            if w in e:
                known_warning = True
                break
        if not known_warning and e != "":
            error_lines += [e]
    return "\n".join(error_lines)


def get_hash():
    h = uuid.uuid4().hex.replace("-", "")
    return h[:5]


def is_tool(name):
    return which(name) is not None


def get_jupyter_bin_path():
    if is_tool("jupyter"):
        return "jupyter"
    if is_tool("jupyter.exe"):
        return "jupyter"
    if sys.executable.endswith("exe"):
        return os.path.join(os.path.dirname(sys.executable), "Scripts", "jupyter.exe")
    return os.path.join(os.path.dirname(sys.executable), "jupyter")


def nb_default_title(nb_path):
    try:
        fname = os.path.basename(nb_path)
        if "." in fname:
            return ".".join(fname.split(".")[:-1])
        return fname
    except Exception as e:
        log.exception("Problem when get default title from notebook")

    return "Please provide title"


def is_presentation(nb):
    for cell in nb.cells:
        if "slideshow" in cell.get("metadata", {}):
            return True
    return False


def make_unique(slug):
    previous_slugs = Notebook.objects.values_list("slug", flat=True)
    if slug not in previous_slugs:
        return slug
    for i in range(1000000):
        slug_unique = f"{slug}-{i}"
        if slug_unique not in previous_slugs:
            return slug_unique

    return slug


def task_init_notebook(
    notebook_path,
    render_html=True,
    is_watch_mode=False,
    notebook_id=None,
    bucket_key=None,
    site=None,
    user=None,
):
    try:
        params = {
            "title": "",
            "author": "Please provide author",
            "description": "Please provide description",
            "output": "app",
            "format": {},
            "schedule": "",
            "notify": {},
        }
        nb = None

        with open(notebook_path, encoding="utf-8", errors="ignore") as f:
            nb = nbformat.read(f, as_version=4)
            parse_params(nb, params)

        if nb is None:
            raise Exception(f"Cant read notebook from {notebook_path}")

        if "date" in params:
            params["date"] = str(params["date"])

        notebook_title = params.get("title", "")
        if notebook_title == "":
            notebook_title = nb_default_title(notebook_path)
        notebook_output = params.get("output", "app")
        notebook_format = params.get("format", {})
        notebook_schedule = params.get("schedule", "")
        notebook_notify = params.get("notify", {})

        if is_presentation(nb):
            # automatically detect slides in cells
            # and set slides output
            notebook_output = "slides"

        if notebook_schedule != "":
            try:
                croniter.is_valid(notebook_schedule)
            except Exception as e:
                raise Exception(
                    f"The schedule ({notebook_schedule}) is not valid. {str(e)} Please check schedule at https://crontab.guru"
                )

        notebook_slug = "some-slug"
        if notebook_id is None:
            notebook_slug = params.get("slug", "")
            if notebook_slug == "":
                fname = os.path.basename(notebook_path)
                fname = fname.replace(".ipynb", "")
                notebook_slug = slugify(fname)
                if notebook_slug is None or notebook_slug == "":
                    notebook_slug = f"nb-{get_hash()}"

            notebook_slug = make_unique(notebook_slug)
        else:
            tmp_nb = Notebook.objects.get(pk=notebook_id)
            notebook_slug = tmp_nb.slug

        notebook_output_file = notebook_slug
        if notebook_id is not None:
            notebook_output_file = f"{notebook_slug}-{get_hash()}"

        if render_html:
            exporter = Exporter(
                show_code=params.get("show-code", False),
                show_prompt=params.get("show-prompt", False),
                is_presentation=notebook_output == "slides",
                reveal_theme=notebook_format.get("theme", "white"),
            )
            body = exporter.export(nb)

            with open(
                os.path.join(settings.MEDIA_ROOT, f"{notebook_output_file}.html"),
                "w",
                encoding="utf-8",
                errors="ignore",
            ) as fout:
                fout.write(body)

            error_msg = ""  # TODO: handle errors

            if not params.get("show-code", False):  # "--no-input" in command:
                with open(
                    os.path.join(settings.MEDIA_ROOT, f"{notebook_output_file}.html"),
                    "a",
                    encoding="utf-8",
                    errors="ignore",
                ) as fout:
                    fout.write(
                        """\n<style type="text/css">
.jp-mod-noOutputs {
    padding: 0px; 
}
.jp-mod-noInput {
  padding-top: 0px;
  padding-bottom: 0px;
}
</style>"""
                    )

        parse_errors = validate_notify(notebook_notify)

        if notebook_id is None:
            if user is None:
                if not User.objects.filter(username="developer"):
                    user = User.objects.create_user(
                        username="developer",
                        email="developer@example.com",
                        password="developer",
                    )
                    EmailAddress.objects.create(
                        user=user, email=user.email, verified=True, primary=True
                    )
                else:
                    user = User.objects.get(username="developer")
            if site is None:
                if not Site.objects.filter(slug="single-site"):
                    site = Site.objects.create(
                        title="Mercury",
                        slug="single-site",
                        share=Site.PUBLIC,
                        created_by=user,
                        status=SiteStatus.READY,
                    )
                else:
                    site = Site.objects.get(slug="single-site")

            bucket_key_fname = ""
            if bucket_key is not None:
                s3 = S3()
                bucket_key_fname = bucket_key.replace(
                    "<replace>", f"{notebook_output_file}.html"
                )
                s3.upload_file(
                    os.path.join(settings.MEDIA_ROOT, f"{notebook_output_file}.html"),
                    bucket_key_fname,
                )

            notebook = Notebook(
                title=notebook_title,
                slug=notebook_slug,
                path=os.path.abspath(notebook_path),
                state="WATCH_READY" if is_watch_mode else "READY",
                params=json.dumps(params),
                default_view_path=bucket_key_fname
                if bucket_key is not None
                else os.path.join(settings.MEDIA_URL, f"{notebook_output_file}.html"),
                file_updated_at=make_aware(
                    datetime.fromtimestamp(os.path.getmtime(notebook_path))
                ),
                output=notebook_output,
                format=json.dumps(notebook_format),
                schedule=notebook_schedule,
                notify=json.dumps(notebook_notify),
                errors=parse_errors,
                created_by=user,
                hosted_on=site,
            )
        else:
            notebook = Notebook.objects.get(pk=notebook_id)
            notebook.title = notebook_title
            notebook.slug = notebook_slug
            notebook.path = os.path.abspath(notebook_path)
            notebook.state = "WATCH_READY" if is_watch_mode else "READY"
            notebook.params = json.dumps(params)
            # remove old default view
            if os.path.exists(notebook.default_view_path):
                os.remove(notebook.default_view_path)
            notebook.default_view_path = os.path.join(
                settings.MEDIA_URL, f"{notebook_output_file}.html"
            )
            notebook.file_updated_at = make_aware(
                datetime.fromtimestamp(os.path.getmtime(notebook_path))
            )
            notebook.output = notebook_output
            notebook.format = json.dumps(notebook_format)
            notebook.schedule = notebook_schedule
            notebook.notify = json.dumps(notebook_notify)
            notebook.errors = parse_errors

        notebook.save()
        return notebook.id
    except KeyboardInterrupt:
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)
    except Exception as e:
        if notebook_id is not None:
            raise e
        else:
            print("Error during notebook initialization.", str(e))
            print(traceback.format_exc())


@shared_task(bind=True)
def task_watch(self, notebook_id):
    notebook = None
    try:
        notebook = Notebook.objects.get(pk=notebook_id)
        current_update_time = datetime.fromtimestamp(os.path.getmtime(notebook.path))
        notebook_updated_at = notebook.file_updated_at.replace(tzinfo=None)

        if current_update_time != notebook_updated_at and notebook.state in [
            "WATCH_READY",
            "WATCH_ERROR",
        ]:
            notebook.state = "WATCH_WAIT"
            notebook.save()
            # clear all tasks
            Task.objects.filter(notebook__id=notebook.id).delete()
            # initialize updated notebook
            task_init_notebook(
                notebook.path,
                render_html=True,
                is_watch_mode=True,
                notebook_id=notebook.id,
            )
    except Exception as e:
        if notebook is not None:
            notebook.state = "WATCH_ERROR"
            notebook.save()



# Generated by Django 3.2.5 on 2023-03-28 08:50

import apps.accounts.fields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("workers", "0001_initial"),
        ("accounts", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="WorkerFile",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("filename", models.CharField(max_length=1024)),
                ("filepath", models.CharField(max_length=1024)),
                ("output_dir", models.CharField(max_length=1024)),
                ("local_filepath", models.CharField(max_length=1024)),
                (
                    "created_at",
                    apps.accounts.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="workers.worker"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UploadedFile",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("filename", models.CharField(max_length=1024)),
                ("filepath", models.CharField(max_length=1024)),
                ("filetype", models.CharField(max_length=128)),
                ("filesize", models.IntegerField()),
                (
                    "created_at",
                    apps.accounts.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "hosted_on",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="accounts.site"
                    ),
                ),
            ],
        ),
    ]



import json
import time
import logging
from datetime import timedelta

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer
from django.conf import settings
from django.db import transaction
from django.db.models import Q
from django.utils import timezone

from apps.accounts.models import Membership, Site
from apps.notebooks.models import Notebook
from apps.workers.models import Worker
from apps.ws.tasks import task_start_websocket_worker
from apps.ws.utils import client_group, worker_group
from apps.storage.s3utils import clean_worker_files

log = logging.getLogger(__name__)


class ClientProxy(WebsocketConsumer):
    def connect(self):
        log.debug("Trying to connect client")

        self.notebook_id = int(self.scope["url_route"]["kwargs"]["notebook_id"])
        self.session_id = self.scope["url_route"]["kwargs"]["session_id"]

        self.user = self.scope["user"]
        nb = Notebook.objects.get(pk=self.notebook_id)
        if nb.hosted_on.share == Site.PRIVATE:
            if self.user.is_anonymous:
                self.close()
            else:
                member = Membership.objects.filter(user=self.user, host=nb.hosted_on)
                owner = nb.hosted_on.created_by == self.user
                if not member and not owner:
                    self.close()

        log.debug(f"Client connect to {self.notebook_id}/{self.session_id}")

        self.client_group = client_group(self.notebook_id, self.session_id)
        self.worker_group = worker_group(self.notebook_id, self.session_id)

        async_to_sync(self.channel_layer.group_add)(
            self.client_group, self.channel_name
        )

        self.server_address = None

        self.start_time = time.time()
        self.site_owner = nb.hosted_on.created_by

        self.accept()

    def disconnect(self, close_code):
        #
        # close worker
        #
        async_to_sync(self.channel_layer.group_send)(
            self.worker_group,
            {"type": "broadcast_message", "payload": {"purpose": "close-worker"}},
        )

        async_to_sync(self.channel_layer.group_discard)(
            self.client_group, self.channel_name
        )
        # log usage
        usage = time.time() - self.start_time
        prev_usage = json.loads(self.site_owner.profile.usage)

        if "usage" in prev_usage:
            prev_usage["usage"] += usage
        else:
            prev_usage["usage"] = usage

        self.site_owner.profile.usage = json.dumps(prev_usage)
        self.site_owner.profile.save()

    def receive(self, text_data):
        log.debug(f"Received from client: {text_data}")

        json_data = json.loads(text_data)

        if json_data.get("purpose", "") == "worker-ping":
            self.worker_ping()
        elif json_data.get("purpose", "") == "server-address":
            self.server_address = json_data.get("address")
            self.need_worker()
        elif json_data.get("purpose", "") == "run-notebook":
            async_to_sync(self.channel_layer.group_send)(
                self.worker_group,
                {"type": "broadcast_message", "payload": json_data},
            )
        elif json_data.get("purpose", "") in [
            "save-notebook",
            "display-notebook",
            "download-html",
            "download-pdf",
        ]:
            async_to_sync(self.channel_layer.group_send)(
                self.worker_group,
                {"type": "broadcast_message", "payload": json_data},
            )

    def broadcast_message(self, event):
        payload = event["payload"]
        self.send(text_data=json.dumps(payload))

    def need_worker(self):
        if self.server_address is None:
            return
        with transaction.atomic():
            log.debug("Create worker in db")
            worker = Worker(
                session_id=self.session_id,
                notebook_id=self.notebook_id,
                state="Queued",
            )
            worker.save()
            job_params = {
                "notebook_id": self.notebook_id,
                "session_id": self.session_id,
                "worker_id": worker.id,
                #
                # ugly hack for docker deployment
                #
                "server_url": self.server_address
                if "0.0.0.0" not in self.server_address
                else self.server_address + ":9000",
            }
            transaction.on_commit(lambda: task_start_websocket_worker.delay(job_params))

    def worker_ping(self):
        workers = Worker.objects.filter(
            Q(state="Running") | Q(state="Queued") | Q(state="Busy"),
            session_id=self.session_id,
            notebook_id=self.notebook_id,
        )

        if not workers:
            self.need_worker()
        else:
            async_to_sync(self.channel_layer.group_send)(
                self.worker_group,
                {
                    "type": "broadcast_message",
                    "payload": {"purpose": "worker-ping"},
                },
            )

        if workers.filter(state="Queued"):
            async_to_sync(self.channel_layer.group_send)(
                self.client_group,
                {
                    "type": "broadcast_message",
                    "payload": {"purpose": "worker-state", "state": "Queued"},
                },
            )

        # clear stale workers
        workers = Worker.objects.filter(
            updated_at__lte=timezone.now()
            - timedelta(minutes=settings.WORKER_STALE_TIME)
        )
        # clean s3 data for worker
        for worker in workers:
            clean_worker_files(worker.notebook.hosted_on.id, worker.session_id)
        
        workers.delete()



from datetime import timedelta

from django.conf import settings
from django.db import transaction
from django.db.models import Q
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404
from django.utils import timezone
from rest_framework import status, viewsets
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.notebooks.models import Notebook
from apps.notebooks.serializers import NotebookSerializer
from apps.workers.models import Worker, WorkerState
from apps.workers.serializers import WorkerSerializer
from apps.storage.s3utils import clean_worker_files

class WorkerGetNb(APIView):
    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            nb = Notebook.objects.get(pk=notebook_id)
            return Response(NotebookSerializer(nb).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class WorkerUpdateNb(APIView):
    def post(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            nb = Notebook.objects.get(pk=notebook_id)
            nb.title = request.data.get("title", nb.title)
            nb.params = request.data.get("params", nb.params)
            nb.save()
            return Response(NotebookSerializer(nb).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class GetWorker(APIView):
    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            return Response(WorkerSerializer(worker).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class IsWorkerStale(APIView):
    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )

            is_stale = worker.updated_at < timezone.now() - timedelta(
                minutes=settings.WORKER_STALE_TIME
            )
            return Response({"is_stale": is_stale})
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class SetWorkerState(APIView):
    def post(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            worker.state = request.data.get("state", WorkerState.Unknown)
            if request.data.get("machine_id") is not None:
                worker.machine_id = request.data.get("machine_id")
            worker.save()
            return Response(WorkerSerializer(worker).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class DeleteWorker(APIView):
    def post(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )

            # delete worker output files
            clean_worker_files(worker.notebook.hosted_on.id, worker.session_id)

            worker.delete()

            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)



import json
import os
import shutil
import sys
import time
import traceback
from re import sub
from shutil import copyfile
from subprocess import PIPE, Popen

import nbformat
from celery import shared_task
from django.conf import settings
from django_drf_filepond.models import TemporaryUpload

from apps.notebooks.models import Notebook
from apps.notebooks.slides_themes import SlidesThemes
from apps.notebooks.tasks import get_jupyter_bin_path, process_nbconvert_errors
from apps.tasks.clean_service import clean_service
from apps.tasks.export_pdf import to_pdf
from apps.tasks.models import Task
from apps.tasks.notify import notify


def get_parameters_cell_index(cells, all_variables):
    max_cnt, max_index = 0, -1
    for i in range(min(len(cells), 10)):
        cnt = 0
        if cells[i]["cell_type"] == "code" and not cells[i]["source"].startswith("---"):
            for v in all_variables:
                if v in cells[i]["source"]:
                    cnt += 1
        if cnt == len(all_variables):
            return i
        if cnt > max_cnt:
            max_cnt = cnt
            max_index = i

    return max_index


def sanitize_string(input_string):
    return sub("""[\"\'(){}\[\]\`\^]""", "", input_string)


@shared_task(bind=True)
def task_execute(self, job_params):
    wrk_input_nb_path = None
    try:
        task = Task.objects.get(pk=job_params["db_id"])
        task.state = "RECEIVED"
        # task.task_id = self.request.get("id", "not-available") # in case of unit tests, id is None
        task.save()

        notebook = Notebook.objects.get(pk=task.notebook_id)

        task_params = {}
        if task.params:
            task_params = json.loads(task.params)

        notebook_params = json.loads(notebook.params)
        widgets_params = notebook_params.get("params", {})

        # validate input variables
        inject_code = ""
        all_variables = []
        remove_after_execution = []
        for k, v in widgets_params.items():
            if v.get("output", "") == "markdown":
                continue
            all_variables += [k]
            use_default = True
            if k in task_params:
                if v["input"] == "text":
                    task_value = task_params[k]
                    inject_code += f'{k} = """{sanitize_string(task_value)}"""\n'
                    use_default = False

                elif v["input"] == "file":
                    file_server_id = task_params[k]
                    # Get the temporary upload record
                    tu = TemporaryUpload.objects.get(upload_id=file_server_id)

                    input_file = f"{task.id}_{tu.upload_name}"
                    input_path = os.path.join(
                        os.path.dirname(notebook.path), input_file
                    )
                    copyfile(tu.get_file_path(), input_path)

                    inject_code += f'{k} = "{input_file}"\n'
                    use_default = False

                    remove_after_execution += [input_path]

                    # DO NOT Delete the temporary upload record and the temporary directory
                    # the file is kept in the UI, maybe user want to reuse it one more time
                    # it will be removed later by cleaning service
                    # DO NOT tu.delete()
                elif v["input"] == "numeric":
                    task_value = task_params[k]
                    allowed_minimum = widgets_params[k].get("min")
                    allowed_maximum = widgets_params[k].get("max")

                    # do not validate the numeric input for minimum and maximum value
                    # validation will be done in the frontend
                    # good_value = True
                    # if allowed_minimum is not None and task_value < allowed_minimum:
                    #    good_value = False
                    # if allowed_maximum is not None and task_value > allowed_maximum:
                    #    good_value = False

                    if (
                        str(task_value)
                        .replace(".", "", 1)
                        .replace(",", "")
                        .replace("-", "", 1)
                        .replace("e", "", 1)
                        .isdigit()
                    ):  # and good_value:
                        inject_code += f"{k} = {task_value}\n"
                        use_default = False
                elif v["input"] == "checkbox":
                    task_value = task_params[k]
                    if task_value in [True, False]:
                        task_value = "True" if task_value else "False"
                        inject_code += f"{k} = {task_value}\n"
                        use_default = False
                elif v["input"] == "select":
                    task_value = task_params[k]
                    isMulti = widgets_params[k].get("multi", False)
                    if not isMulti and task_value in widgets_params[k]["choices"]:
                        inject_code += f'{k} = "{task_value}"\n'
                        use_default = False
                    if isMulti:
                        values_str_list = []
                        for t in task_value:
                            if t in widgets_params[k]["choices"]:
                                values_str_list += [f'"{t}"']
                        if values_str_list or len(task_value) == 0:
                            inject_code += f'{k} = [{",".join(values_str_list)}]\n'
                            use_default = False
                elif v["input"] == "slider":
                    task_value = task_params[k]

                    if (
                        (
                            str(task_value)
                            .replace(".", "", 1)
                            .replace(",", "")
                            .replace("-", "", 1)
                            .replace("e", "", 1)
                            .isdigit()
                        )
                        and (task_value >= widgets_params[k].get("min", 0))
                        and (task_value <= widgets_params[k].get("max", 100))
                    ):
                        inject_code += f"{k} = {task_value}\n"
                        use_default = False
                elif v["input"] == "range":
                    task_value = task_params[k]

                    if (
                        len(task_value) == 2
                        and (
                            str(task_value[0])
                            .replace(".", "", 1)
                            .replace(",", "")
                            .replace("-", "", 1)
                            .replace("e", "", 1)
                            .isdigit()
                        )
                        and (
                            str(task_value[1])
                            .replace(".", "", 1)
                            .replace(",", "")
                            .replace("-", "", 1)
                            .replace("e", "", 1)
                            .isdigit()
                        )
                        and (task_value[0] >= widgets_params[k].get("min", 0))
                        and (task_value[0] <= widgets_params[k].get("max", 100))
                        and (task_value[1] >= widgets_params[k].get("min", 0))
                        and (task_value[1] <= widgets_params[k].get("max", 100))
                    ):
                        inject_code += f"{k} = {task_value}\n"
                        use_default = False

            if use_default:
                if widgets_params[k].get("value") is not None:
                    if v["input"] in ["text", "file", "select"]:
                        inject_code += f'{k} = "{widgets_params[k].get("value", "")}"\n'
                    else:
                        inject_code += f'{k} = {widgets_params[k].get("value")}\n'

        # create output directory for notebook output
        wrk_dir = settings.MEDIA_ROOT / task.session_id
        if not os.path.exists(wrk_dir):
            try:
                os.mkdir(wrk_dir)
            except Exception as e:
                raise Exception(f"Cant create {wrk_dir}")

        # validate output variables
        for k, v in widgets_params.items():
            if v.get("output") is not None and v.get("output") == "dir":
                # create output directory for files created in the notebook
                output_dir = settings.MEDIA_ROOT / task.session_id / f"output_{task.id}"
                if not os.path.exists(output_dir):
                    try:
                        os.mkdir(output_dir)
                    except Exception as e:
                        raise Exception(f"Cant create {output_dir}")
                # pass path to directory into the notebook's code
                inject_code += f'{k} = r"{str(output_dir)}"\n'

            if v.get("output") is not None and v.get("output") == "response":
                # create output directory for REST response created in the notebook
                output_dir = settings.MEDIA_ROOT / task.session_id
                if not os.path.exists(output_dir):
                    try:
                        os.mkdir(output_dir)
                    except Exception as e:
                        raise Exception(f"Cant create {output_dir}")
                # pass path to directory into the notebook's code
                output_dir = output_dir / "response.json"
                inject_code += f'{k} = r"{str(output_dir)}"\n'

        new_cell = {
            "cell_type": "code",
            "execution_count": None,
            # "id": "1234",
            "metadata": {},
            "outputs": [],
            "source": inject_code,
        }

        # update input notebook with params from the task
        # the input notebook path should be the same as original notebook
        # so it can access data with relative paths
        wrk_input_nb_path = os.path.join(
            os.path.dirname(notebook.path), f"input_{task.id}_{task.session_id}.ipynb"
        )
        wrk_output_nb_file = f"output_{task.id}.html"

        if all_variables:
            with open(notebook.path, encoding="utf-8", errors="ignore") as f:
                nb = nbformat.read(f, as_version=4)
                if "cells" in nb and len(nb["cells"]) > 0:
                    old_parameters_index = get_parameters_cell_index(
                        nb["cells"], all_variables
                    )
                    if old_parameters_index != -1 and old_parameters_index < len(
                        nb["cells"]
                    ):
                        new_cell["metadata"] = nb["cells"][old_parameters_index].get(
                            "metadata", {}
                        )
                        del nb["cells"][old_parameters_index]
                        nb["cells"].insert(
                            old_parameters_index, nbformat.from_dict(new_cell)
                        )
                    else:
                        nb["cells"].insert(0, nbformat.from_dict(new_cell))

                    with open(
                        wrk_input_nb_path, "w", encoding="utf-8", errors="ignore"
                    ) as f:
                        nbformat.write(nb, f)
        else:
            shutil.copyfile(notebook.path, wrk_input_nb_path)

        command = [
            get_jupyter_bin_path(),
            "nbconvert",
            "--RegexRemovePreprocessor.patterns='^---'",
            "--execute",
            "--allow-errors",
            str(wrk_input_nb_path),
            "--output",
            wrk_output_nb_file,
            "--output-dir",
            str(wrk_dir),
            "--to",
            "slides" if notebook.output == "slides" else "html",
        ]
        if "show-code" in notebook_params and not notebook_params["show-code"]:
            command += ["--no-input"]
        if "show-prompt" in notebook_params and not notebook_params["show-prompt"]:
            command += ["--no-prompt"]

        if notebook.output == "slides":
            command += SlidesThemes.nbconvert_options(json.loads(notebook.format))

        error_msg = ""
        with Popen(command, stdout=PIPE, stderr=PIPE) as proc:
            # print(proc.stdout.read())
            # print(proc.stderr.read())
            error_msg = proc.stderr.read()

        error_msg = process_nbconvert_errors(error_msg)

        # check if output file exists
        output_html_path = os.path.join(str(wrk_dir), wrk_output_nb_file)

        # change file name if needed
        if notebook.output == "slides":
            slides_fpath = os.path.join(
                str(wrk_dir), f"{wrk_output_nb_file}.slides.html"
            )
            if os.path.exists(slides_fpath):
                os.rename(slides_fpath, output_html_path)

        if os.path.isfile(output_html_path):
            if "--no-input" in command:
                with open(
                    wrk_dir / wrk_output_nb_file, "a", encoding="utf-8", errors="ignore"
                ) as fout:
                    fout.write(
                        """\n<style type="text/css">
.jp-mod-noOutputs {
    padding: 0px; 
}
.jp-mod-noInput {
  padding-top: 0px;
  padding-bottom: 0px;
}
</style>"""
                    )

            if notebook.output == "slides":
                with open(
                    wrk_dir / wrk_output_nb_file, "a", encoding="utf-8", errors="ignore"
                ) as fout:
                    fout.write(SlidesThemes.additional_css(json.loads(notebook.format)))

            task.result = f"{settings.MEDIA_URL}{task.session_id}/{wrk_output_nb_file}"
            task.state = "DONE"
        else:
            task.result = (
                error_msg if error_msg != "" else "Problem with executing the notebook"
            )
            task.state = "ERROR"
        task.save()
    except Exception as e:
        task.result = str(e)
        task.state = "ERROR"
        task.save()
        print(str(e))
        print(traceback.format_exc())
    finally:
        # notify
        print(notebook.schedule)
        print(notebook.notify)

        if notebook.schedule:
            notify(
                json.loads(notebook.notify),
                is_success=task.state == "DONE",
                error_msg="" if task.state == "DONE" else task.result,
                notebook_id=notebook.id,
                notebook_url=task.result,
            )

        # remove input notebook
        if wrk_input_nb_path is not None:
            if os.path.isfile(wrk_input_nb_path):
                os.remove(wrk_input_nb_path)
        # remove copied files from temporary uploads
        for f in remove_after_execution:
            if os.path.exists(f):
                os.remove(f)
        # remove old file
        clean_service()



# Please run tests with below command
# python manage.py test apps.accounts
import os
from datetime import datetime

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.core import mail
from django.utils.timezone import make_aware
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from apps.accounts.models import Membership, Secret, Site
from apps.notebooks.models import Notebook
from apps.workers.models import Worker


class SitesTestCase(APITestCase):
    register_url = "/api/v1/auth/register/"
    verify_email_url = "/api/v1/auth/register/verify-email/"
    login_url = "/api/v1/auth/login/"
    user_details_url = "/api/v1/auth/user/"
    logout_url = "/api/v1/auth/logout/"
    delete_account_url = "/api/v1/auth/delete-account/"
    change_password_url = "/api/v1/auth/password/change/"
    reset_password_url = "/api/v1/auth/password/reset/"
    reset_password_confirm_url = "/api/v1/auth/password/reset/confirm/"

    sites_url = "/api/v1/sites/"
    members_url = "/api/v1/{}/members/"  # need to set site_id
    invite_url = "/api/v1/{}/invite/"

    def setUp(self):
        #
        # first user
        #
        self.user1_params = {
            "username": "user1",  # it is optional to pass username
            "email": "piotr@example.com",
            "password": "verysecret",
        }
        self.user = User.objects.create_user(
            username=self.user1_params["username"],
            email=self.user1_params["email"],
            password=self.user1_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        #
        # second user
        #
        self.user2_params = {
            "username": "user2",  # it is optional to pass username
            "email": "piotr2@example.com",
            "password": "verysecret2",
        }
        self.user2 = User.objects.create_user(
            username=self.user2_params["username"],
            email=self.user2_params["email"],
            password=self.user2_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user2, email=self.user2.email, verified=True, primary=True
        )

    def test_list_sites(self):
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )
        # login as first user to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 1)

        # login as second user
        response = self.client.post(self.login_url, self.user2_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 0)
        # add rights to edit
        Membership.objects.create(
            user=self.user2, host=site, rights=Membership.EDIT, created_by=self.user
        )
        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 1)

    def test_get_site(self):
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )
        # login to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        response = self.client.get(f"{self.sites_url}1/", **headers)
        for key in ["title", "slug", "share", "created_by", "created_at"]:
            self.assertTrue(key in response.json())

    def test_create_site(self):
        # login to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        # no sites in database
        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 0)
        # create site
        new_data = {"title": "New site", "slug": "Some slug", "share": Site.PRIVATE}
        response = self.client.post(self.sites_url, new_data, **headers)
        self.assertEqual(response.status_code, 201)
        for key in ["title", "slug", "share", "created_by", "created_at"]:
            self.assertTrue(key in response.json())

    def test_destroy_site(self):
        # create site for first user
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )
        # login as second user to get token
        response = self.client.post(self.login_url, self.user2_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        # second user cant delete
        response = self.client.delete(f"{self.sites_url}1/", **headers)
        self.assertEqual(response.status_code, 404)

        # login as first user to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 1)

        # first user delete site because only owner can delete
        response = self.client.delete(f"{self.sites_url}1/", **headers)
        self.assertEqual(response.status_code, 204)

        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 0)

    def test_update_site(self):
        # create site for first user
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )

        # login as second user to get token
        response = self.client.post(self.login_url, self.user2_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        new_data = {"title": "New title", "slug": "new-slug", "share": Site.PRIVATE}
        # try to edit without edit rights
        response = self.client.put(f"{self.sites_url}1/", new_data, **headers)
        self.assertEqual(response.status_code, 404)

        # give edit access to second user
        Membership.objects.create(
            user=self.user2, host=site, rights=Membership.EDIT, created_by=self.user
        )
        # second user can edit
        response = self.client.put(f"{self.sites_url}1/", new_data, **headers)
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data["title"], new_data["title"])
        self.assertEqual(data["slug"], new_data["slug"])
        self.assertEqual(data["share"], new_data["share"])

    def test_update_site_invalid_slug(self):
        # create site for first user
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )

        # login as second user to get token
        response = self.client.post(self.login_url, self.user2_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        new_data = {"title": "New title", "slug": "mljar", "share": Site.PRIVATE}
        # try to edit without edit rights
        response = self.client.put(f"{self.sites_url}1/", new_data, **headers)
        self.assertEqual(response.status_code, 404)

        # give edit access to second user
        Membership.objects.create(
            user=self.user2, host=site, rights=Membership.EDIT, created_by=self.user
        )
        # second user can edit
        response = self.client.put(f"{self.sites_url}1/", new_data, **headers)
        self.assertEqual(response.status_code, 400)
        self.assertTrue("You cant" in response.json()[0])

        # second user can edit
        response = self.client.patch(
            f"{self.sites_url}1/", {"slug": "mljar"}, **headers
        )
        self.assertEqual(response.status_code, 400)
        self.assertTrue("You cant" in response.json()[0])

        # second user can edit
        response = self.client.patch(
            f"{self.sites_url}1/", {"title": "new title"}, **headers
        )
        self.assertEqual(response.status_code, 200)

    def test_limit_sites(self):
        # login to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        # no sites in database
        response = self.client.get(self.sites_url, **headers)
        self.assertEqual(len(response.json()), 0)
        # create site
        new_data = {"title": "New site", "slug": "Some slug", "share": Site.PRIVATE}
        response = self.client.post(self.sites_url, new_data, **headers)
        self.assertEqual(response.status_code, 201)

        os.environ["MERCURY_CLOUD"] = "1"
        # create site, but limit should be reached
        new_data = {"title": "New site", "slug": "Some slug 2", "share": Site.PRIVATE}
        response = self.client.post(self.sites_url, new_data, **headers)
        self.assertEqual(response.status_code, 403)

        os.environ["MERCURY_CLOUD"] = "0"
        # create site, now it should work
        new_data = {
            "title": "New site",
            "slug": "Some slug 2 asd",
            "share": Site.PRIVATE,
        }
        response = self.client.post(self.sites_url, new_data, **headers)
        print(response.content)
        self.assertEqual(response.status_code, 201)

        # create site with the same slug
        new_data = {"title": "New site", "slug": "Some slug", "share": Site.PRIVATE}
        response = self.client.post(self.sites_url, new_data, **headers)
        self.assertEqual(response.status_code, 403)

        # create site with the forbidden slug
        new_data = {"title": "New site", "slug": "mercury", "share": Site.PRIVATE}
        response = self.client.post(self.sites_url, new_data, **headers)
        self.assertEqual(response.status_code, 403)



import json
import os
import tempfile

import ipywidgets
from IPython.display import display

from .manager import WidgetsManager


class File:
    def __init__(self, label="File upload", max_file_size="100MB"):
        self.max_file_size = max_file_size
        self.code_uid = WidgetsManager.get_code_uid("File")

        if WidgetsManager.widget_exists(self.code_uid):
            self.file = WidgetsManager.get_widget(self.code_uid)
            self.file.description = label
        else:
            self.file = ipywidgets.FileUpload(description=label)
            self.file.filepath = None
            self.file.filename = None
            WidgetsManager.add_widget(self.file.model_id, self.code_uid, self.file)
        display(self)

    @property
    def value(self):
        if len(self.file.value):
            return self.file.value[0].content

        if self.file.filepath is not None:
            # read that file
            with open(
                self.file.filepath, "rb", encoding="utf-8", errors="ignore"
            ) as fin:
                return fin.read()

        return None

    @property
    def filename(self):
        if len(self.file.value):
            return self.file.value[0].name
        if self.file.filename is not None:
            return self.file.filename
        return None

    @property
    def filepath(self):
        if self.file.filepath is not None:
            return self.file.filepath

        if (
            len(self.file.value)
            and self.filename is not None
            and self.value is not None
        ):
            # store file in temp dir
            # and return the path
            temp_dir = tempfile.TemporaryDirectory()
            self.file.filepath = os.path.join(temp_dir.name, self.filename)

            with open(self.file.filepath, "wb") as fout:
                fout.write(self.value)

            return self.file.filepath

        return None

    @value.setter
    def value(self, v):
        filename, filepath = v
        self.file.filepath = filepath
        self.file.filename = filename

    def __str__(self):
        return "mercury.File"

    def __repr__(self):
        return "mercury.File"

    def _repr_mimebundle_(self, **kwargs):
        # data = {}
        # data["text/plain"] = repr(self)
        # return data
        data = self.file._repr_mimebundle_()

        if len(data) > 1:
            view = {
                "widget": "File",
                "max_file_size": self.max_file_size,
                "label": self.file.description,
                "model_id": self.file.model_id,
                "code_uid": self.code_uid,
            }
            data["application/mercury+json"] = json.dumps(view, indent=4)
            if "text/plain" in data:
                del data["text/plain"]

            return data



import os
import json
import uuid
from IPython.display import display_javascript, display_html, display, HTML

JSON_CSS = """.renderjson a              { text-decoration: none; }
.renderjson .disclosure    { color: grey; font-size: 125%; }
.renderjson .syntax        { color: grey; }
.renderjson .string        { color: #fe46a5; }
.renderjson .number        { color: #0f9b8e; }
.renderjson .boolean       { color: black; }
.renderjson .key           { color: #2684ff; }
.renderjson .keyword       { color: gray; }
.renderjson .object.syntax { color: gray; }
.renderjson .array.syntax  { color: gray; }"""


def JSON(json_data={}, level=1):
    if isinstance(json_data, dict):
        json_str = json.dumps(json_data)
    else:
        json_str = json_data
    id = str(uuid.uuid4())

    # Minimized version based on https://raw.githubusercontent.com/caldwell/renderjson/master/renderjson.js
    # Changes:
    # 1. Replace \n with \\n
    # 2. Remove module exports
    js = """var renderjson=function(){var t=function(){for(var t=[];arguments.length;)t.push(n(s(Array.prototype.shift.call(arguments)),o(Array.prototype.shift.call(arguments))));return t},n=function(){for(var t=Array.prototype.shift.call(arguments),e=0;e<arguments.length;e++)arguments[e].constructor==Array?n.apply(this,[t].concat(arguments[e])):t.appendChild(arguments[e]);return t},e=function(t,n){return t.insertBefore(n,t.firstChild),t},r=function(t,n){var e=n||Object.keys(t);for(var r in e)if(Object.hasOwnProperty.call(t,e[r]))return!1;return!0},o=function(t){return document.createTextNode(t)},s=function(t){var n=document.createElement("span");return t&&(n.className=t),n},l=function(t,n,e){var r=document.createElement("a");return n&&(r.className=n),r.appendChild(o(t)),r.href="#",r.onclick=function(t){return e(),t&&t.stopPropagation(),!1},r};function a(i,c,u,p,y){var _=u?"":c,f=function(r,a,i,c,u){var f,g=s(c),h=function(){f||n(g.parentNode,f=e(u(),l(y.hide,"disclosure",(function(){f.style.display="none",g.style.display="inline"})))),f.style.display="inline",g.style.display="none"};n(g,l(y.show,"disclosure",h),t(c+" syntax",r),l(a,null,h),t(c+" syntax",i));var d=n(s(),o(_.slice(0,-1)),g);return p>0&&"string"!=c&&h(),d};return null===i?t(null,_,"keyword","null"):void 0===i?t(null,_,"keyword","undefined"):"string"==typeof i&&i.length>y.max_string_length?f('"',i.substr(0,y.max_string_length)+" ...",'"',"string",(function(){return n(s("string"),t(null,_,"string",JSON.stringify(i)))})):"object"!=typeof i||[Number,String,Boolean,Date].indexOf(i.constructor)>=0?t(null,_,typeof i,JSON.stringify(i)):i.constructor==Array?0==i.length?t(null,_,"array syntax","[]"):f("[",y.collapse_msg(i.length),"]","array",(function(){for(var e=n(s("array"),t("array syntax","[",null,"\\n")),r=0;r<i.length;r++)n(e,a(y.replacer.call(i,r,i[r]),c+"    ",!1,p-1,y),r!=i.length-1?t("syntax",","):[],o("\\n"));return n(e,t(null,c,"array syntax","]")),e})):r(i,y.property_list)?t(null,_,"object syntax","{}"):f("{",y.collapse_msg(Object.keys(i).length),"}","object",(function(){var e=n(s("object"),t("object syntax","{",null,"\\n"));for(var r in i)var l=r;var u=y.property_list||Object.keys(i);for(var _ in y.sort_objects&&(u=u.sort()),u){(r=u[_])in i&&n(e,t(null,c+"    ","key",'"'+r+'"',"object syntax",": "),a(y.replacer.call(i,r,i[r]),c+"    ",!0,p-1,y),r!=l?t("syntax",","):[],o("\\n"))}return n(e,t(null,c,"object syntax","}")),e}))}var i=function t(e){var r=new Object(t.options);r.replacer="function"==typeof r.replacer?r.replacer:function(t,n){return n};var o=n(document.createElement("pre"),a(e,"",!1,r.show_to_level,r));return o.className="renderjson",o};return i.set_icons=function(t,n){return i.options.show=t,i.options.hide=n,i},i.set_show_to_level=function(t){return i.options.show_to_level="string"==typeof t&&"all"===t.toLowerCase()?Number.MAX_VALUE:t,i},i.set_max_string_length=function(t){return i.options.max_string_length="string"==typeof t&&"none"===t.toLowerCase()?Number.MAX_VALUE:t,i},i.set_sort_objects=function(t){return i.options.sort_objects=t,i},i.set_replacer=function(t){return i.options.replacer=t,i},i.set_collapse_msg=function(t){return i.options.collapse_msg=t,i},i.set_property_list=function(t){return i.options.property_list=t,i},i.set_show_by_default=function(t){return i.options.show_to_level=t?Number.MAX_VALUE:0,i},i.options={},i.set_icons("⊕","⊖"),i.set_show_by_default(!1),i.set_sort_objects(!1),i.set_max_string_length("none"),i.set_replacer(void 0),i.set_property_list(void 0),i.set_collapse_msg((function(t){return t+" item"+(1==t?"":"s")})),i}();"""

    display(HTML(f'<style>{JSON_CSS}</style><div id="{id}"></div>'))
    display(
        HTML(
            f'<script>{js} renderjson.set_show_to_level({str(level)}); document.getElementById("{id}").appendChild(renderjson({json_str}))</script>'
        )
    )



from django.conf import settings
from django.db import transaction
from django.db.models import Q
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404
from rest_framework import status, viewsets
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.accounts.models import Membership, Site
from apps.notebooks.models import Notebook
from apps.notebooks.serializers import NotebookSerializer
from apps.notebooks.tasks import task_init_notebook, task_watch
from apps.storage.s3utils import S3


def in_commas(word):
    return "," + word + ","


def notebooks_queryset(request, site_id):
    user = request.user
    if user.is_anonymous:
        return Notebook.objects.filter(
            hosted_on__id=site_id, hosted_on__share=Site.PUBLIC
        )

    # it can be optimized
    site = Site.objects.get(pk=site_id)
    if site.share == Site.PUBLIC:
        return Notebook.objects.filter(hosted_on=site)

    # admin can see all notebooks on site
    if site.created_by == user:
        return Notebook.objects.filter(hosted_on=site)

    # don't filter on rights because both VIEW and EDIT allows
    # to see and execute notebooks
    m = Membership.objects.filter(user=user, host=site)
    if m:
        return Notebook.objects.filter(hosted_on=site)
    return Notebook.objects.filter(hosted_on=site, created_by=user)


def nb_iframe_url(request, iframe_db_address):
    iframe_url = ""
    if settings.STORAGE == settings.STORAGE_MEDIA:
        host = request.build_absolute_uri("/")[:-1]
        iframe_url = f"{host}{iframe_db_address}"
    elif settings.STORAGE == settings.STORAGE_S3:
        s3 = S3()
        url = s3.get_presigned_url(iframe_db_address, "get_object")
        iframe_url = url
    return iframe_url


class ListNotebooks(APIView):
    def get(self, request, site_id, format=None):
        notebooks = notebooks_queryset(request, site_id).order_by("slug")

        # build iframe urls
        for n in notebooks:
            n.default_view_path = nb_iframe_url(request, n.default_view_path)

        serializer = NotebookSerializer(notebooks, many=True)
        return JsonResponse(serializer.data, safe=False)


class RetrieveNotebook(APIView):
    def get(self, request, site_id, notebook_id, format=None):
        pk = int(notebook_id.replace("/", ""))
        notebook = get_object_or_404(notebooks_queryset(request, site_id), pk=pk)
        notebook.default_view_path = nb_iframe_url(request, notebook.default_view_path)

        serializer = NotebookSerializer(notebook)
        if notebook.state.startswith("WATCH"):
            task_watch.delay(notebook.id)

        return JsonResponse(serializer.data, safe=False)


class RetrieveNotebookWithSlug(APIView):
    def get(self, request, site_id, notebook_slug, format=None):
        notebook_slug = notebook_slug.replace("/", "")

        notebooks = notebooks_queryset(request, site_id).filter(slug=notebook_slug)

        if not notebooks:
            return JsonResponse({}, status=status.HTTP_404_NOT_FOUND)

        notebooks[0].default_view_path = nb_iframe_url(
            request, notebooks[0].default_view_path
        )

        # get the first one - it should be only one :)
        serializer = NotebookSerializer(notebooks[0])
        return JsonResponse(serializer.data, safe=False)


class GetNbIframes(APIView):
    def get(self, request, site_id, format=None):
        site = Site.objects.get(pk=site_id)

        notebooks = notebooks_queryset(request, site_id)

        urls = {}
        s3 = S3()
        for n in notebooks:
            url = s3.get_presigned_url(n.default_view_path, "get_object")
            urls[n.slug] = url

        return Response(urls)



from datetime import datetime, timedelta

from django.utils.timezone import make_aware
from django_drf_filepond.models import TemporaryUpload


def clean_service():
    # clean temporary uploads older than 1 day
    # delete in DB remove files from storage as well
    tus = TemporaryUpload.objects.filter(
        uploaded__lte=make_aware(datetime.now() - timedelta(days=1))
    )
    tus.delete()



from django.db import models
from django.utils.timezone import now


class AutoCreatedField(models.DateTimeField):
    """
    A DateTimeField that automatically populates itself at
    object creation.
    By default, sets editable=False, default=datetime.now.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", False)
        kwargs.setdefault("default", now)
        super(AutoCreatedField, self).__init__(*args, **kwargs)


class AutoLastModifiedField(AutoCreatedField):
    """
    A DateTimeField that updates itself on each save() of the model.
    By default, sets editable=False and default=datetime.now.
    """

    def pre_save(self, model_instance, add):
        value = now()
        setattr(model_instance, self.attname, value)
        return value



from django.urls import re_path

from apps.ws.client import ClientProxy
from apps.ws.worker import WorkerProxy

websocket_urlpatterns = [
    re_path(
        r"ws/client/(?P<notebook_id>.+)/(?P<session_id>.+)/$",
        ClientProxy.as_asgi(),
    ),
    re_path(
        r"ws/worker/(?P<notebook_id>.+)/(?P<session_id>.+)/(?P<worker_id>.+)/$",
        WorkerProxy.as_asgi(),
    ),
]



import os
import uuid
import inspect
import logging

log = logging.getLogger(__name__)


mercury_widgets = {}
# widgets_mapping = {}
widget_index_to_model_id = {}
widgets_counter = 0
cell_index_to_widgets_index = {}


class WidgetException(Exception):
    pass


class WidgetsManager:
    widgets = {}  # model_id -> widget
    code2model = {}  # code generated uid  -> model_id
    cell_index = 0  # current cell index

    @staticmethod
    def rand_uid():
        if os.environ.get("RUN_MERCURY") is None:
            h = uuid.uuid4().hex.replace("-", "")
            return f"-rand{h[:8]}"
        return ""

    @staticmethod
    def set_cell_index(new_index):
        WidgetsManager.cell_index = new_index

    @staticmethod
    def get_code_uid(widget_type="widget", key=""):
        uid = f"{widget_type}.{WidgetsManager.cell_index}"
        for frame in inspect.stack():
            info = inspect.getframeinfo(frame[0])
            if info.function == "run_code":
                break
            uid += f".{info.lineno}"
        if key != "":
            uid += f".{key}"
        uid += WidgetsManager.rand_uid()
        return uid

    @staticmethod
    def fix_cell_index(code_uid, correct_cell_index):
        # remove rand uid
        code_uid = code_uid.split("-rand")[0]
        # fix cell index
        parts = code_uid.split(".")
        parts[1] = f"{correct_cell_index}"
        return ".".join(parts)

    @staticmethod
    def parse_cell_index(code_uid):
        return int(code_uid.split(".")[1])

    @staticmethod
    def parse_widget_type(code_uid):
        return code_uid.split(".")[0]

    @staticmethod
    def widget_exists(code_uid):
        return code_uid in WidgetsManager.code2model

    @staticmethod
    def get_widget(code_uid):
        model_id = WidgetsManager.code2model.get(code_uid)
        if model_id is None:
            return None
        return WidgetsManager.widgets.get(model_id)

    @staticmethod
    def add_widget(model_id, code_uid, widget):
        WidgetsManager.widgets[model_id] = widget
        WidgetsManager.code2model[code_uid] = model_id

    @staticmethod
    def update(code_uid, field, new_value):
        # returns
        # True if there was update
        # False if no update
        model_id = WidgetsManager.code2model.get(code_uid)
        if model_id is None:
            return False
        w = WidgetsManager.widgets.get(model_id)
        if w is not None:
            if getattr(w, field) != new_value:
                setattr(w, field, new_value)
                return True

        return False

    @staticmethod
    def frontend_format(output):
        widget_type = output.get("widget", "unknown")

        if widget_type == "Slider":
            return {
                "input": "slider",
                "value": output.get("value", 0),
                "min": output.get("min", 0),
                "max": output.get("max", 10),
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "Select":
            return {
                "input": "select",
                "value": output.get("value", ""),
                "choices": output.get("choices", []),
                "multi": False,
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "MultiSelect":
            return {
                "input": "select",
                "value": output.get("value", []),
                "choices": output.get("choices", []),
                "multi": True,
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "Range":
            return {
                "input": "range",
                "value": output.get("value", [0, 1]),
                "min": output.get("min", 0),
                "max": output.get("max", 10),
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "Text":
            return {
                "input": "text",
                "value": output.get("value", ""),
                "rows": output.get("rows", 1),
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "File":
            return {
                "input": "file",
                "maxFileSize": output.get("max_file_size", 1),
                "label": output.get("label", ""),
            }
        elif widget_type == "OutputDir":
            return {
                "output": "dir",
                "value": "",
            }
        elif widget_type == "Checkbox":
            return {
                "input": "checkbox",
                "value": output.get("value", True),
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "Numeric":
            return {
                "input": "numeric",
                "value": output.get("value", 0),
                "min": output.get("min", 0),
                "max": output.get("max", 10),
                "step": output.get("step", 1),
                "label": output.get("label", ""),
                "url_key": output.get("url_key", ""),
            }
        elif widget_type == "Note":
            return {
                "output": "markdown",
                "value": output.get("value", ""),
            }
        elif widget_type == "Button":
            return {
                "input": "button",
                "value": output.get("value", False),
                "label": output.get("label", ""),
                "style": output.get("style", "primary"),
            }

        return {}




import json
import os
import sys

import nbformat
from execnb.nbio import nb2dict, nb2str, read_nb, write_nb
from execnb.shell import CaptureShell

CURRENT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BACKEND_DIR = os.path.join(CURRENT_DIR, "..")
sys.path.insert(0, BACKEND_DIR)

import copy

from execnb.nbio import _dict2obj, dict2nb

from apps.nb.exporter import Exporter
from apps.nb.utils import one_cell_notebook, test_notebook


class NbRun:
    def __init__(
        self,
        show_code=False,
        show_prompt=False,
        is_presentation=False,
        reveal_theme="white",
    ):
        self.exporter = Exporter(show_code, show_prompt, is_presentation, reveal_theme)
        self.shell = CaptureShell()
        try:
            self.shell.enable_matplotlib("inline")
        except Exception as e:
            pass
        self.shell.run("from mercury import WidgetsManager")
        self.shell.run("import os\nos.environ['RUN_MERCURY']='1'")

    def set_show_code(self, new_show_code):
        self.exporter.set_show_code(new_show_code)

    def set_show_code_and_prompt(self, new_show_code, new_show_prompt):
        self.exporter.set_show_prompt(new_show_prompt)
        self.exporter.set_show_code(new_show_code)

    def set_is_presentation(self, new_value):
        self.exporter.set_is_presentation(new_value)

    def set_reveal_theme(self, new_value):
        self.exporter.set_reveal_theme(new_value)

    def run_set_cell_index(self, new_index):
        if new_index is not None:
            self.shell.run(f"WidgetsManager.set_cell_index({new_index})")

    def run_code(self, code):
        return self.shell.run(code)

    def run_cell(self, cell, counter=None):
        if cell.cell_type == "code":
            self.run_set_cell_index(counter)

            cell.outputs = []
            self.shell.cell(cell)

            if counter is not None:
                cell.execution_count = counter

            for output in cell.outputs:
                if output.get(
                    "output_type", ""
                ) == "error" and "StopExecution" in output.get("ename", ""):
                    return False
        return True

    def run_notebook(self, nb, start=0):
        #
        # nb is fastai format
        #
        counter = start + 1
        continue_computing = True
        for c in nb.cells[start:]:
            if continue_computing:
                continue_computing = self.run_cell(c, counter)
                counter += 1
            else:
                if c.cell_type == "code":
                    c.outputs = []

    def export_html(self, nb, full_header=True):
        body = self.exporter.export(
            nbformat.reads(nb2str(nb), as_version=4), full_header
        )

        return body

    def get_header(self):
        nb = one_cell_notebook("print(1)")
        nb = dict2nb(nb)
        e = NbRun()
        body = e.run_notebook(nb)
        index_start = body.find("<head>")
        index_end = body.find("</head>")

        if index_start != -1 and index_end != -1:
            return body[index_start : index_end + 7]
        return ""


# run it to fill cache
# other runs will be faster
nb = one_cell_notebook("print(1)")
nb = dict2nb(nb)
e = NbRun()
e.run_notebook(nb)

# NbRun = NbRun(is_presentation=True)
# nb_original = read_nb("./slides.ipynb")
# b = NbRun.run_notebook(nb_original, export_html=True)

# print(nb_original)
# print(b)
# with open("test.html", "w") as fout:
#    fout.write(b)

# nb = get_test_notebook(code=["import handcalcs.render", "\n%%render\na=1"])
# nb = dict2nb(nb)
# e = NbRun()
# b = e.run_notebook(nb)

# # print(b)
# with open("test.html", "w") as fout:
#    fout.write(b)


# nb.cells += [
#     _dict2obj({
#         "cell_type": "code",
#         "execution_count": None,
#         "id": "7fasdfasdf",
#         "metadata": {},
#         "outputs": [],
#         "source": 'import datetime\nprint(datetime.datetime.now())\n',
#         "idx_": 3,
#     })
# ]



import os
import json
import requests
from rest_framework import permissions, status
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.serializers import ValidationError

from apps.accounts.serializers import UserSerializer

MERCURY_CLOUD_PRO_ID = 820856
MERCURY_CLOUD_BUSINESS_ID = 820858

PADDLE_VENDOR_ID = os.environ.get("PADDLE_VENDOR_ID")
PADDLE_API_KEY = os.environ.get("PADDLE_API_KEY")


class SubscriptionView(APIView):
    permission_classes = [permissions.IsAuthenticated]

    def check_subscription(self, request):
        checkout_id = request.data.get("checkoutId")
        if checkout_id is None:
            raise ValidationError("Missing checkout id")

        url = f"https://checkout.paddle.com/api/1.0/order?checkout_id={checkout_id}"
        print(url)
        response = requests.get(url)
        user = request.user
        if response.status_code != 200:
            user.profile.info = json.dumps(
                {
                    "plan": "starter",
                    "status": "error",
                    "error": "Problem with subscription check",
                }
            )
            user.profile.save()
            raise ValidationError("Cant receive data from paddle")

        data = response.json()
        subscription_id = data.get("order", {}).get("subscription_id")
        receipt_url = data.get("order", {}).get("receipt_url")
        product_id = data.get("order", {}).get("product_id")

        print(data)

        if subscription_id is None:
            user.profile.info = json.dumps(
                {
                    "plan": "starter",
                    "status": "error",
                    "error": "Problem with subscription check, empty Paddle reponse",
                }
            )
        else:
            user.profile.info = json.dumps(
                {
                    "plan": "business"
                    if product_id == MERCURY_CLOUD_BUSINESS_ID
                    else "pro",
                    "status": "active",
                    "error": "",
                    "subscription_id": subscription_id,
                    "receipt_url": receipt_url,
                    "product_id": product_id,
                }
            )
        user.profile.save()

    def is_active(self, request):
        user = request.user
        user_data = json.loads(user.profile.info)

        plan = user_data.get("plan", "starter")

        if plan != "starter":
            if PADDLE_API_KEY is None or PADDLE_VENDOR_ID is None:
                raise ValidationError("Paddle not configured")

            url = "https://vendors.paddle.com/api/2.0/subscription/users"

            subscription_id = user_data.get("subscription_id")
            if subscription_id is None:
                raise ValidationError("Missing user subscription information")

            data = {
                "vendor_id": PADDLE_VENDOR_ID,
                "vendor_auth_code": PADDLE_API_KEY,
                "subscription_id": subscription_id,
                "state": "active",
            }
            response = requests.post(url, data=data)
            if response.status_code != 200:
                raise ValidationError("Error when get information from Paddle")

            success = response.json().get("success", False)
            data = response.json().get("response", [])

            if not success or len(data) == 0:
                user_data["plan"] = "starter"
                user_data["status"] = "not_active"

            else:
                user_data["cancel_url"] = data[0].get("cancel_url", "")
                user_data["status"] = "active"

            user.profile.info = json.dumps(user_data)
            user.profile.save()

    def post(self, request, format=None):
        action = request.data.get("action")

        if action == "check":
            self.check_subscription(request)
            self.is_active(request)
        elif action == "is_active":
            self.is_active(request)

        user = request.user

        return Response(UserSerializer(user).data, status=status.HTTP_200_OK)



from django.contrib import admin

# Register your models here.
from .models import Task


class TaskModelAdmin(admin.ModelAdmin):
    list_display = ("id", "session_id", "notebook", "state")

    class Meta:
        model = Task


admin.site.register(Task, TaskModelAdmin)



from rest_framework import permissions

from apps.accounts.models import Membership, Site


class HasEditRights(permissions.BasePermission):
    def has_permission(self, request, view):
        site_id = view.kwargs.get("site_id")
        if site_id is None:  # just in case
            return False
        return (
            Membership.objects.filter(
                host__id=site_id, user=request.user, rights=Membership.EDIT
            )
            or Site.objects.get(pk=site_id).created_by == request.user
        )

    def has_object_permission(self, request, view, obj):
        return (
            Membership.objects.filter(
                host=obj.host, user=request.user, rights=Membership.EDIT
            )
            or obj.host.created_by == request.user
        )
